import os

import pytest
import torch

import ignite.distributed as idist
from ignite.engine import Engine, Events
from ignite.handlers import EarlyStopping


def do_nothing_update_fn(engine, batch):
    pass


def test_args_validation():
    trainer = Engine(do_nothing_update_fn)

    with pytest.raises(ValueError, match=r"Argument patience should be positive integer."):
        EarlyStopping(patience=-1, score_function=lambda engine: 0, trainer=trainer)

    with pytest.raises(ValueError, match=r"Argument min_delta should not be a negative number."):
        EarlyStopping(patience=2, min_delta=-0.1, score_function=lambda engine: 0, trainer=trainer)

    with pytest.raises(TypeError, match=r"Argument score_function should be a function."):
        EarlyStopping(patience=2, score_function=12345, trainer=trainer)

    with pytest.raises(TypeError, match=r"Argument trainer should be an instance of Engine."):
        EarlyStopping(patience=2, score_function=lambda engine: 0, trainer=None)


def test_simple_early_stopping():
    scores = iter([1.0, 0.8, 0.88])

    def score_function(engine):
        return next(scores)

    trainer = Engine(do_nothing_update_fn)

    h = EarlyStopping(patience=2, score_function=score_function, trainer=trainer)
    # Call 3 times and check if stopped
    assert not trainer.should_terminate
    h(None)
    assert not trainer.should_terminate
    h(None)
    assert not trainer.should_terminate
    h(None)
    assert trainer.should_terminate


def test_state_dict():
    scores = iter([1.0, 0.8, 0.88])

    def score_function(engine):
        return next(scores)

    trainer = Engine(do_nothing_update_fn)

    h = EarlyStopping(patience=2, score_function=score_function, trainer=trainer)
    # Call 3 times and check if stopped
    assert not trainer.should_terminate
    h(None)
    assert not trainer.should_terminate

    # Swap to new object, but maintain state
    h2 = EarlyStopping(patience=2, score_function=score_function, trainer=trainer)
    h2.load_state_dict(h.state_dict())

    h2(None)
    assert not trainer.should_terminate
    h2(None)
    assert trainer.should_terminate


def test_early_stopping_on_delta():
    scores = iter([1.0, 2.0, 2.01, 3.0, 3.01, 3.02])

    trainer = Engine(do_nothing_update_fn)

    h = EarlyStopping(patience=2, min_delta=0.1, score_function=lambda _: next(scores), trainer=trainer)

    assert not trainer.should_terminate
    h(None)  # counter == 0
    assert not trainer.should_terminate
    h(None)  # delta == 1.0; counter == 0
    assert not trainer.should_terminate
    h(None)  # delta == 0.01; counter == 1
    assert not trainer.should_terminate
    h(None)  # delta == 0.99; counter == 0
    assert not trainer.should_terminate
    h(None)  # delta == 0.01; counter == 1
    assert not trainer.should_terminate
    h(None)  # delta == 0.01; counter == 2
    assert trainer.should_terminate


def test_early_stopping_on_last_event_delta():
    scores = iter([0.0, 0.3, 0.6])

    trainer = Engine(do_nothing_update_fn)

    h = EarlyStopping(
        patience=2, min_delta=0.4, cumulative_delta=False, score_function=lambda _: next(scores), trainer=trainer
    )

    assert not trainer.should_terminate
    h(None)  # counter == 0
    assert not trainer.should_terminate
    h(None)  # delta == 0.3; counter == 1
    assert not trainer.should_terminate
    h(None)  # delta == 0.3; counter == 2
    assert trainer.should_terminate


def test_early_stopping_on_cumulative_delta():
    scores = iter([0.0, 0.3, 0.6])

    trainer = Engine(do_nothing_update_fn)

    h = EarlyStopping(
        patience=2, min_delta=0.4, cumulative_delta=True, score_function=lambda _: next(scores), trainer=trainer
    )

    assert not trainer.should_terminate
    h(None)  # counter == 0
    assert not trainer.should_terminate
    h(None)  # delta == 0.3; counter == 1
    assert not trainer.should_terminate
    h(None)  # delta == 0.6; counter == 0
    assert not trainer.should_terminate


def test_simple_early_stopping_on_plateau():
    def score_function(engine):
        return 42

    trainer = Engine(do_nothing_update_fn)

    h = EarlyStopping(patience=1, score_function=score_function, trainer=trainer)
    # Call 2 times and check if stopped
    assert not trainer.should_terminate
    h(None)
    assert not trainer.should_terminate
    h(None)
    assert trainer.should_terminate


def test_simple_no_early_stopping():
    scores = iter([1.0, 0.8, 1.2])

    def score_function(engine):
        return next(scores)

    trainer = Engine(do_nothing_update_fn)

    h = EarlyStopping(patience=2, score_function=score_function, trainer=trainer)
    # Call 3 times and check if not stopped
    assert not trainer.should_terminate
    h(None)
    h(None)
    h(None)
    assert not trainer.should_terminate


def test_with_engine_early_stopping():
    class Counter(object):
        def __init__(self, count=0):
            self.count = count

    n_epochs_counter = Counter()

    scores = iter([1.0, 0.8, 1.2, 1.5, 0.9, 1.0, 0.99, 1.1, 0.9])

    def score_function(engine):
        return next(scores)

    trainer = Engine(do_nothing_update_fn)
    evaluator = Engine(do_nothing_update_fn)
    early_stopping = EarlyStopping(patience=3, score_function=score_function, trainer=trainer)

    @trainer.on(Events.EPOCH_COMPLETED)
    def evaluation(engine):
        evaluator.run([0])
        n_epochs_counter.count += 1

    evaluator.add_event_handler(Events.COMPLETED, early_stopping)
    trainer.run([0], max_epochs=10)
    assert n_epochs_counter.count == 7
    assert trainer.state.epoch == 7


def test_with_engine_early_stopping_on_plateau():
    class Counter(object):
        def __init__(self, count=0):
            self.count = count

    n_epochs_counter = Counter()

    def score_function(engine):
        return 0.047

    trainer = Engine(do_nothing_update_fn)
    evaluator = Engine(do_nothing_update_fn)
    early_stopping = EarlyStopping(patience=4, score_function=score_function, trainer=trainer)

    @trainer.on(Events.EPOCH_COMPLETED)
    def evaluation(engine):
        evaluator.run([0])
        n_epochs_counter.count += 1

    evaluator.add_event_handler(Events.COMPLETED, early_stopping)
    trainer.run([0], max_epochs=10)
    assert n_epochs_counter.count == 5
    assert trainer.state.epoch == 5


def test_with_engine_no_early_stopping():
    class Counter(object):
        def __init__(self, count=0):
            self.count = count

    n_epochs_counter = Counter()

    scores = iter([1.0, 0.8, 1.2, 1.23, 0.9, 1.0, 1.1, 1.253, 1.26, 1.2])

    def score_function(engine):
        return next(scores)

    trainer = Engine(do_nothing_update_fn)
    evaluator = Engine(do_nothing_update_fn)
    early_stopping = EarlyStopping(patience=5, score_function=score_function, trainer=trainer)

    @trainer.on(Events.EPOCH_COMPLETED)
    def evaluation(engine):
        evaluator.run([0])
        n_epochs_counter.count += 1

    evaluator.add_event_handler(Events.COMPLETED, early_stopping)
    trainer.run([0], max_epochs=10)
    assert n_epochs_counter.count == 10
    assert trainer.state.epoch == 10


def _test_distrib_with_engine_early_stopping(device):
    if device is None:
        device = idist.device()
    if isinstance(device, str):
        device = torch.device(device)

    torch.manual_seed(12)

    class Counter(object):
        def __init__(self, count=0):
            self.count = count

    n_epochs_counter = Counter()

    scores = torch.tensor([1.0, 0.8, 1.2, 1.5, 0.9, 1.0, 0.99, 1.1, 0.9], requires_grad=False).to(device)

    def score_function(engine):
        i = trainer.state.epoch - 1
        v = scores[i]
        idist.all_reduce(v)
        v /= idist.get_world_size()
        return v.item()

    trainer = Engine(do_nothing_update_fn)
    evaluator = Engine(do_nothing_update_fn)
    early_stopping = EarlyStopping(patience=3, score_function=score_function, trainer=trainer)

    @trainer.on(Events.EPOCH_COMPLETED)
    def evaluation(engine):
        evaluator.run([0])
        n_epochs_counter.count += 1

    evaluator.add_event_handler(Events.COMPLETED, early_stopping)
    trainer.run([0], max_epochs=10)
    assert trainer.state.epoch == 7
    assert n_epochs_counter.count == 7


def _test_distrib_integration_engine_early_stopping(device):
    from ignite.metrics import Accuracy

    if device is None:
        device = idist.device()
    if isinstance(device, str):
        device = torch.device(device)
    metric_device = device
    if device.type == "xla":
        metric_device = "cpu"

    rank = idist.get_rank()
    ws = idist.get_world_size()
    torch.manual_seed(12)

    n_epochs = 10
    n_iters = 20

    y_preds = (
        [torch.randint(0, 2, size=(n_iters, ws)).to(device)]
        + [torch.ones(n_iters, ws).to(device)]
        + [torch.randint(0, 2, size=(n_iters, ws)).to(device) for _ in range(n_epochs - 2)]
    )

    y_true = (
        [torch.randint(0, 2, size=(n_iters, ws)).to(device)]
        + [torch.ones(n_iters, ws).to(device)]
        + [torch.randint(0, 2, size=(n_iters, ws)).to(device) for _ in range(n_epochs - 2)]
    )

    def update(engine, _):
        e = trainer.state.epoch - 1
        i = engine.state.iteration - 1
        return y_preds[e][i, rank], y_true[e][i, rank]

    evaluator = Engine(update)
    acc = Accuracy(device=metric_device)
    acc.attach(evaluator, "acc")

    def score_function(engine):
        return engine.state.metrics["acc"]

    trainer = Engine(lambda e, b: None)
    early_stopping = EarlyStopping(patience=3, score_function=score_function, trainer=trainer)

    @trainer.on(Events.EPOCH_COMPLETED)
    def evaluation(engine):
        data = list(range(n_iters))
        evaluator.run(data=data)

    evaluator.add_event_handler(Events.COMPLETED, early_stopping)
    trainer.run([0], max_epochs=10)
    assert trainer.state.epoch == 5


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl):
    device = idist.device()
    _test_distrib_with_engine_early_stopping(device)
    _test_distrib_integration_engine_early_stopping(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):
    device = idist.device()
    _test_distrib_with_engine_early_stopping(device)
    _test_distrib_integration_engine_early_stopping(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor):
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_distrib_with_engine_early_stopping, (device,), np=nproc, do_init=True)
    gloo_hvd_executor(_test_distrib_integration_engine_early_stopping, (device,), np=nproc, do_init=True)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_gloo_cpu_or_gpu(distributed_context_multi_node_gloo):
    device = idist.device()
    _test_distrib_with_engine_early_stopping(device)
    _test_distrib_integration_engine_early_stopping(device)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_nccl_gpu(distributed_context_multi_node_nccl):
    device = idist.device()
    _test_distrib_with_engine_early_stopping(device)
    _test_distrib_integration_engine_early_stopping(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_single_device_xla():
    device = idist.device()
    _test_distrib_with_engine_early_stopping(device)
    _test_distrib_integration_engine_early_stopping(device)


def _test_distrib_xla_nprocs(index):
    device = idist.device()
    _test_distrib_with_engine_early_stopping(device)
    _test_distrib_integration_engine_early_stopping(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



import logging
import re
from unittest.mock import MagicMock

import pytest
import torch
import torch.nn as nn
import torch.optim as optim

from ignite.engine import create_supervised_trainer, Engine, Events
from ignite.handlers.fbresearch_logger import FBResearchLogger
from ignite.utils import setup_logger


@pytest.fixture
def mock_engine():
    engine = Engine(lambda e, b: None)
    engine.state.epoch = 1
    engine.state.max_epochs = 10
    engine.state.epoch_length = 100
    engine.state.iteration = 50
    return engine


@pytest.fixture
def mock_logger():
    return MagicMock(spec=logging.Logger)


@pytest.fixture
def fb_research_logger(mock_logger):
    yield FBResearchLogger(logger=mock_logger, show_output=True)


def test_fbresearch_logger_initialization(mock_logger):
    logger = FBResearchLogger(logger=mock_logger, show_output=True)
    assert logger.logger == mock_logger
    assert logger.show_output is True


def test_fbresearch_logger_attach(mock_engine, mock_logger):
    logger = FBResearchLogger(logger=mock_logger, show_output=True)
    logger.attach(mock_engine, name="Test", every=1)
    assert mock_engine.has_event_handler(logger.log_every, Events.ITERATION_COMPLETED)


@pytest.mark.parametrize(
    "output,expected_pattern",
    [
        ({"loss": 0.456, "accuracy": 0.789}, r"loss. *0.456.*accuracy. *0.789"),
        ((0.456, 0.789), r"0.456.*0.789"),
        ([0.456, 0.789], r"0.456.*0.789"),
    ],
)
def test_output_formatting(mock_engine, fb_research_logger, output, expected_pattern):
    # Ensure the logger correctly formats and logs the output for each type
    mock_engine.state.output = output
    fb_research_logger.attach(mock_engine, name="Test", every=1)
    mock_engine.fire_event(Events.ITERATION_COMPLETED)

    actual_output = fb_research_logger.logger.info.call_args_list[0].args[0]
    assert re.search(expected_pattern, actual_output)


def test_logger_type_support():
    model = nn.Linear(10, 5)
    opt = optim.SGD(model.parameters(), lr=0.001)
    criterion = nn.CrossEntropyLoss()

    data = [(torch.rand(4, 10), torch.randint(0, 5, size=(4,))) for _ in range(100)]

    trainer = create_supervised_trainer(model, opt, criterion)

    logger = setup_logger("trainer", level=logging.INFO)
    logger = FBResearchLogger(logger=logger, show_output=True)
    logger.attach(trainer, name="Train", every=20, optimizer=opt)

    trainer.run(data, max_epochs=4)
    trainer.state.output = {"loss": 4.2}
    trainer.fire_event(Events.ITERATION_COMPLETED)
    trainer.state.output = "4.2"
    trainer.fire_event(Events.ITERATION_COMPLETED)
    trainer.state.output = [4.2, 4.2]
    trainer.fire_event(Events.ITERATION_COMPLETED)
    trainer.state.output = (4.2, 4.2)
    trainer.fire_event(Events.ITERATION_COMPLETED)


def test_fbrlogger_with_output_transform(mock_logger):
    trainer = Engine(lambda e, b: 42)
    fbr = FBResearchLogger(logger=mock_logger, show_output=True)
    fbr.attach(trainer, "Training", output_transform=lambda x: {"loss": x})
    trainer.run(data=[10], epoch_length=1, max_epochs=1)
    assert "loss: 42.0000" in fbr.logger.info.call_args_list[-2].args[0]


def test_fbrlogger_with_state_attrs(mock_logger):
    trainer = Engine(lambda e, b: 42)
    fbr = FBResearchLogger(logger=mock_logger, show_output=True)
    fbr.attach(trainer, "Training", state_attributes=["alpha", "beta", "gamma"])
    trainer.state.alpha = 3.899
    trainer.state.beta = torch.tensor(12.21)
    trainer.state.gamma = torch.tensor([21.0, 6.0])
    trainer.run(data=[10], epoch_length=1, max_epochs=1)
    attrs = "alpha: 3.8990 beta: 12.2100 gamma: [21.0000, 6.0000]"
    assert attrs in fbr.logger.info.call_args_list[-2].args[0]



import os

import numpy as np
import pytest
import torch

import ignite.distributed as idist
from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics.regression import MedianAbsolutePercentageError


def test_zero_sample():
    m = MedianAbsolutePercentageError()
    with pytest.raises(
        NotComputableError, match=r"EpochMetric must have at least one example before it can be computed"
    ):
        m.compute()


def test_wrong_input_shapes():
    m = MedianAbsolutePercentageError()

    with pytest.raises(ValueError, match=r"Predictions should be of shape"):
        m.update((torch.rand(4, 1, 2), torch.rand(4, 1)))

    with pytest.raises(ValueError, match=r"Targets should be of shape"):
        m.update((torch.rand(4, 1), torch.rand(4, 1, 2)))

    with pytest.raises(ValueError, match=r"Predictions should be of shape"):
        m.update((torch.rand(4, 1, 2), torch.rand(4)))

    with pytest.raises(ValueError, match=r"Targets should be of shape"):
        m.update((torch.rand(4), torch.rand(4, 1, 2)))


def test_median_absolute_percentage_error(available_device):
    # See https://github.com/torch/torch7/pull/182
    # For even number of elements, PyTorch returns middle element
    # NumPy returns average of middle elements
    # Size of dataset will be odd for these tests

    size = 51  # odd size ensures consistent median behavior

    y_pred = torch.rand(size)
    y = torch.rand(size)

    m = MedianAbsolutePercentageError(device=available_device)
    assert m._device == torch.device(available_device)

    m.reset()
    m.update((y_pred, y))

    # Compute expected result with torch
    abs_perc_errors = 100.0 * torch.abs(y - y_pred) / torch.abs(y)
    expected = torch.median(abs_perc_errors).item()

    assert pytest.approx(expected) == m.compute()


def test_median_absolute_percentage_error_2(available_device):
    torch.manual_seed(1)
    size = 105
    y_pred = torch.rand(size, 1)
    y = torch.rand(size, 1)

    # Shuffle y (like np.random.shuffle)
    indices = torch.randperm(size)
    y = y[indices]

    # Compute expected result using torch
    abs_perc_errors = 100.0 * torch.abs(y - y_pred) / torch.abs(y)
    expected = torch.median(abs_perc_errors).item()

    m = MedianAbsolutePercentageError(device=available_device)
    assert m._device == torch.device(available_device)

    m.reset()
    batch_size = 16
    n_iters = size // batch_size + 1
    for i in range(n_iters):
        idx = i * batch_size
        m.update((y_pred[idx : idx + batch_size], y[idx : idx + batch_size]))

    assert pytest.approx(expected) == m.compute()


def test_integration_median_absolute_percentage_error(available_device):
    torch.manual_seed(1)
    size = 105
    y_pred = torch.rand(size, 1)
    y = torch.rand(size, 1)

    # Shuffle y (similar to np.random.shuffle)
    indices = torch.randperm(size)
    y = y[indices]

    # Compute expected median absolute percentage error using torch
    abs_perc_errors = 100.0 * torch.abs(y - y_pred) / torch.abs(y)
    expected = torch.median(abs_perc_errors).item()

    batch_size = 15

    def update_fn(engine, batch):
        idx = (engine.state.iteration - 1) * batch_size
        return y_pred[idx : idx + batch_size], y[idx : idx + batch_size]

    engine = Engine(update_fn)

    m = MedianAbsolutePercentageError(device=available_device)
    assert m._device == torch.device(available_device)

    m.attach(engine, "median_absolute_percentage_error")

    data = list(range(size // batch_size))
    result = engine.run(data, max_epochs=1).metrics["median_absolute_percentage_error"]

    assert pytest.approx(expected) == result


def _test_distrib_compute(device):
    def _test(metric_device):
        metric_device = torch.device(metric_device)
        m = MedianAbsolutePercentageError(device=metric_device)

        size = 105

        y_pred = torch.randint(1, 10, size=(size, 1), dtype=torch.double, device=device)
        y = torch.randint(1, 10, size=(size, 1), dtype=torch.double, device=device)

        m.update((y_pred, y))

        # gather y_pred, y
        y_pred = idist.all_gather(y_pred)
        y = idist.all_gather(y)

        np_y_pred = y_pred.cpu().numpy().ravel()
        np_y = y.cpu().numpy().ravel()

        res = m.compute()

        e = np.abs(np_y - np_y_pred) / np.abs(np_y)

        np_res = 100.0 * np.median(e)
        assert pytest.approx(res) == np_res

    rank = idist.get_rank()
    for i in range(3):
        torch.manual_seed(10 + rank + i)
        _test("cpu")
        if device.type != "xla":
            _test(idist.device())


def _test_distrib_integration(device):
    def _test(n_epochs, metric_device):
        metric_device = torch.device(metric_device)
        n_iters = 80
        size = 105
        y_true = torch.rand(size=(n_iters * size,)).to(device)
        y_preds = torch.rand(size=(n_iters * size,)).to(device)

        def update(engine, i):
            return (
                y_preds[i * size : (i + 1) * size],
                y_true[i * size : (i + 1) * size],
            )

        engine = Engine(update)

        m = MedianAbsolutePercentageError(device=metric_device)
        m.attach(engine, "mape")

        data = list(range(n_iters))
        engine.run(data=data, max_epochs=n_epochs)

        y_preds = idist.all_gather(y_preds)
        y_true = idist.all_gather(y_true)

        assert "mape" in engine.state.metrics

        res = engine.state.metrics["mape"]

        np_y_true = y_true.cpu().numpy().ravel()
        np_y_preds = y_preds.cpu().numpy().ravel()

        e = np.abs(np_y_true - np_y_preds) / np.abs(np_y_true)
        np_res = 100.0 * np.median(e)

        assert pytest.approx(res) == np_res

    metric_devices = ["cpu"]
    if device.type != "xla":
        metric_devices.append(idist.device())
    for metric_device in metric_devices:
        rank = idist.get_rank()
        for i in range(2):
            torch.manual_seed(12 + rank + i)
            _test(n_epochs=1, metric_device=metric_device)
            _test(n_epochs=2, metric_device=metric_device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor):
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_distrib_compute, (device,), np=nproc, do_init=True)
    gloo_hvd_executor(_test_distrib_integration, (device,), np=nproc, do_init=True)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_gloo_cpu_or_gpu(distributed_context_multi_node_gloo):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_nccl_gpu(distributed_context_multi_node_nccl):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_single_device_xla():
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


def _test_distrib_xla_nprocs(index):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



from typing import Callable, cast, Dict, List, Optional, Sequence, Tuple, Union

import torch
from typing_extensions import Literal

from ignite.metrics import MetricGroup

from ignite.metrics.mean_average_precision import _BaseAveragePrecision, _cat_and_agg_tensors
from ignite.metrics.metric import Metric, reinit__is_reduced, sync_all_reduce


def coco_tensor_list_to_dict_list(
    output: Tuple[
        Union[List[torch.Tensor], List[Dict[str, torch.Tensor]]],
        Union[List[torch.Tensor], List[Dict[str, torch.Tensor]]],
    ]
) -> Tuple[List[Dict[str, torch.Tensor]], List[Dict[str, torch.Tensor]]]:
    """Convert either of output's `y_pred` or `y` from list of `(N, 6)` tensors to list of str-to-tensor dictionaries,
    or keep them unchanged if they're already in the deisred format.

    Input format is a `(N, 6)` or (`N, 5)` tensor which `N` is the number of predicted/target bounding boxes for the
    image and the second dimension contains `(x1, y1, x2, y2, confidence, class)`/`(x1, y1, x2, y2, class[, iscrowd])`.
    Output format is a str-to-tensor dictionary containing 'bbox' and `class` keys, plus `confidence` key for `y_pred`
    and possibly `iscrowd` for `y`.

    Args:
        output: `(y_pred,y)` tuple whose members are either list of tensors or list of dicts.

    Returns:
        `(y_pred,y)` tuple whose members are list of str-to-tensor dictionaries.
    """
    y_pred, y = output
    if len(y_pred) > 0 and isinstance(y_pred[0], torch.Tensor):
        y_pred = [{"bbox": t[:, :4], "confidence": t[:, 4], "class": t[:, 5]} for t in cast(List[torch.Tensor], y_pred)]
    if len(y) > 0 and isinstance(y[0], torch.Tensor):
        if y[0].size(1) == 5:
            y = [{"bbox": t[:, :4], "class": t[:, 4]} for t in cast(List[torch.Tensor], y)]
        else:
            y = [{"bbox": t[:, :4], "class": t[:, 4], "iscrowd": t[:, 5]} for t in cast(List[torch.Tensor], y)]
    return cast(Tuple[List[Dict[str, torch.Tensor]], List[Dict[str, torch.Tensor]]], (y_pred, y))


class ObjectDetectionAvgPrecisionRecall(Metric, _BaseAveragePrecision):
    _tps: List[torch.Tensor]
    _fps: List[torch.Tensor]
    _scores: List[torch.Tensor]
    _y_pred_labels: List[torch.Tensor]
    _y_true_count: torch.Tensor
    _num_classes: int

    def __init__(
        self,
        iou_thresholds: Optional[Union[Sequence[float], torch.Tensor]] = None,
        rec_thresholds: Optional[Union[Sequence[float], torch.Tensor]] = None,
        num_classes: int = 80,
        max_detections_per_image_per_class: int = 100,
        area_range: Literal["small", "medium", "large", "all"] = "all",
        output_transform: Callable = lambda x: x,
        device: Union[str, torch.device] = torch.device("cpu"),
        skip_unrolling: bool = False,
    ) -> None:
        r"""Calculate mean average precision & recall for evaluating an object detector in the COCO way.


        Average precision is computed by averaging precision over increasing levels of recall thresholds.
        In COCO, the maximum precision across thresholds greater or equal a recall threshold is
        considered as the average summation operand; In other words, the precision peek across lower or equal
        sensivity levels is used for a recall threshold:

        .. math::
            \text{Average Precision} = \sum_{k=1}^{\#rec\_thresholds} (r_k - r_{k-1}) max(P_{k:})

        Average recall is the detector's maximum recall, considering all matched detections as TP,
        averaged over classes.

        Args:
            iou_thresholds: sequence of IoU thresholds to be considered for computing mean average precision & recall.
                Values should be between 0 and 1. If not given, COCO's default (.5, .55, ..., .95) would be used.
            rec_thresholds: sequence of recall thresholds to be considered for computing mean average precision.
                Values should be between 0 and 1. If not given, COCO's default (.0, .01, .02, ..., 1.) would be used.
            num_classes: number of categories. Default is 80, that of the COCO dataset.
            area_range: area range which only objects therein are considered in evaluation. By default, 'all'.
            max_detections_per_image_per_class: maximum number of detections per class in each image to consider
                for evaluation. The most confident ones are selected.
            output_transform: a callable that is used to transform the :class:`~ignite.engine.engine.Engine`'s
                ``process_function``'s output into the form expected by the metric. An already provided example is
                :func:`~ignite.metrics.vision.object_detection_average_precision_recall.coco_tensor_list_to_dict_list`
                which accepts `y_pred` and `y` as lists of tensors and transforms them to the expected format.
                Default is the identity function.
            device: specifies which device updates are accumulated on. Setting the
                metric's device to be the same as your ``update`` arguments ensures the ``update`` method is
                non-blocking. By default, CPU.
            skip_unrolling: specifies whether output should be unrolled before being fed to update method. Should be
                true for multi-output model, for example, if ``y_pred`` and ``y`` contain multi-ouput as
                ``(y_pred_a, y_pred_b)`` and ``(y_a, y_b)``, in which case the update method is called for
                ``(y_pred_a, y_a)`` and ``(y_pred_b, y_b)``.Alternatively, ``output_transform`` can be used to handle
                this.

        .. versionadded:: 0.5.2
        """
        try:
            from torchvision.ops.boxes import _box_inter_union, box_area

            def box_iou(pred_boxes: torch.Tensor, gt_boxes: torch.Tensor, iscrowd: torch.BoolTensor) -> torch.Tensor:
                inter, union = _box_inter_union(pred_boxes, gt_boxes)
                union[:, iscrowd] = box_area(pred_boxes).reshape(-1, 1)
                iou = inter / union
                iou[iou.isnan()] = 0
                return iou

            self.box_iou = box_iou
        except ImportError:
            raise ModuleNotFoundError("This metric requires torchvision to be installed.")

        if iou_thresholds is None:
            iou_thresholds = torch.linspace(0.5, 0.95, 10, dtype=torch.double)

        self._iou_thresholds = self._setup_thresholds(iou_thresholds, "iou_thresholds")

        if rec_thresholds is None:
            rec_thresholds = torch.linspace(0, 1, 101, dtype=torch.double)

        self._num_classes = num_classes
        self._area_range = area_range
        self._max_detections_per_image_per_class = max_detections_per_image_per_class

        super(ObjectDetectionAvgPrecisionRecall, self).__init__(
            output_transform=output_transform,
            device=device,
            skip_unrolling=skip_unrolling,
        )
        super(Metric, self).__init__(
            rec_thresholds=rec_thresholds,
            class_mean=None,
        )
        precision = torch.double if torch.device(device).type != "mps" else torch.float32
        self.rec_thresholds = cast(torch.Tensor, self.rec_thresholds).to(device=device, dtype=precision)

    @reinit__is_reduced
    def reset(self) -> None:
        self._tps = []
        self._fps = []
        self._scores = []
        self._y_pred_labels = []
        self._y_true_count = torch.zeros((self._num_classes,), device=self._device)

    def _match_area_range(self, bboxes: torch.Tensor) -> torch.Tensor:
        from torchvision.ops.boxes import box_area

        areas = box_area(bboxes)
        if self._area_range == "all":
            min_area = 0
            max_area = 1e10
        elif self._area_range == "small":
            min_area = 0
            max_area = 1024
        elif self._area_range == "medium":
            min_area = 1024
            max_area = 9216
        elif self._area_range == "large":
            min_area = 9216
            max_area = 1e10
        return torch.logical_and(areas >= min_area, areas <= max_area)

    def _check_matching_input(
        self, output: Tuple[List[Dict[str, torch.Tensor]], List[Dict[str, torch.Tensor]]]
    ) -> None:
        y_pred, y = output
        if len(y_pred) != len(y):
            raise ValueError(f"y_pred and y should have the same number of samples, given {len(y_pred)} and {len(y)}.")
        if len(y_pred) == 0:
            raise ValueError("y_pred and y should contain at least one sample.")

        y_pred_keys = {"bbox", "scores", "labels"}
        if (y_pred[0].keys() & y_pred_keys) != y_pred_keys:
            raise ValueError(
                "y_pred sample dictionaries should have 'bbox', 'scores'"
                f" and 'labels' keys, given keys: {y_pred[0].keys()}"
            )

        y_keys = {"bbox", "labels"}
        if (y[0].keys() & y_keys) != y_keys:
            raise ValueError(
                "y sample dictionaries should have 'bbox', 'labels'"
                f" and optionally 'iscrowd' keys, given keys: {y[0].keys()}"
            )

    def _compute_recall_and_precision(
        self, TP: torch.Tensor, FP: torch.Tensor, scores: torch.Tensor, y_true_count: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Measuring recall & precision

        This method is different from that of MeanAveragePrecision since in the pycocotools reference implementation,
        when there are predictions with the same scores, they're considered associated with different thresholds in the
        course of measuring recall values, although it's not logically correct as those predictions are really
        associated with a single threshold, thus outputing a single recall value.

        Shape of function inputs and return values follow the table below. N\ :sub:`pred` is the number of detections
        or predictions. ``...`` stands for the possible additional dimensions. Finally, \#unique scores represents
        number of unique scores in ``scores`` which is actually the number of thresholds.

        ============== ======================
        **Object**     **Shape**
        ============== ======================
        TP and FP      (..., N\ :sub:`pred`)
        scores         (N\ :sub:`pred`,)
        y_true_count   () (A single float,
                       greater than zero)
        recall         (..., \#unique scores)
        precision      (..., \#unique scores)
        ============== ======================

        Returns:
            `(recall, precision)`
        """
        indices = torch.argsort(scores, dim=-1, stable=True, descending=True)
        tp = TP[..., indices]
        tp_summation = tp.cumsum(dim=-1)
        if tp_summation.device.type != "mps":
            tp_summation = tp_summation.double()

        fp = FP[..., indices]
        fp_summation = fp.cumsum(dim=-1)
        if fp_summation.device.type != "mps":
            fp_summation = fp_summation.double()

        recall = tp_summation / y_true_count
        predicted_positive = tp_summation + fp_summation
        precision = tp_summation / torch.where(predicted_positive == 0, 1, predicted_positive)

        return recall, precision

    def _compute_average_precision(self, recall: torch.Tensor, precision: torch.Tensor) -> torch.Tensor:
        """Measuring average precision.
        This method is overriden since :math:`1/#recall_thresholds` is used instead of :math:`r_k - r_{k-1}`
        as the recall differential in COCO's reference implementation i.e., pycocotools.

        Args:
            recall: n-dimensional tensor whose last dimension is the dimension of the samples. Should be ordered in
                ascending order in its last dimension.
            precision: like ``recall`` in the shape.

        Returns:
            average_precision: (n-1)-dimensional tensor containing the average precision for mean dimensions.
        """
        if precision.device.type == "mps":
            # Manual fallback to CPU if precision is on MPS due to the error:
            # NotImplementedError: The operator 'aten::_cummax_helper' is not currently implemented for the MPS device
            device = precision.device
            precision_integrand = precision.flip(-1).cpu()
            precision_integrand = precision_integrand.cummax(dim=-1).values
            precision_integrand = precision_integrand.to(device=device).flip(-1)
        else:
            precision_integrand = precision.flip(-1).cummax(dim=-1).values.flip(-1)
        rec_thresholds = cast(torch.Tensor, self.rec_thresholds).repeat((*recall.shape[:-1], 1))
        rec_thresh_indices = (
            torch.searchsorted(recall, rec_thresholds)
            if recall.size(-1) != 0
            else torch.LongTensor([], device=self._device)
        )
        precision_integrand = precision_integrand.take_along_dim(
            rec_thresh_indices.where(rec_thresh_indices != recall.size(-1), 0), dim=-1
        ).where(rec_thresh_indices != recall.size(-1), 0)
        return torch.sum(precision_integrand, dim=-1) / len(cast(torch.Tensor, self.rec_thresholds))

    @reinit__is_reduced
    def update(self, output: Tuple[List[Dict[str, torch.Tensor]], List[Dict[str, torch.Tensor]]]) -> None:
        r"""Metric update method using prediction and target.

        Args:
            output: a tuple, (y_pred, y), of two same-length lists, each one containing
                str-to-tensor dictionaries whose items is as follows. N\ :sub:`det` and
                N\ :sub:`gt` are number of detections and ground truths for a sample
                respectively.

                ======== ================== =================================================
                **y_pred items**
                -----------------------------------------------------------------------------
                Key      Value shape        Description
                ======== ================== =================================================
                'bbox'   (N\ :sub:`det`, 4) Bounding boxes of form (x1, y1, x2, y2)
                                            containing top left and bottom right coordinates.
                'scores' (N\ :sub:`det`,)   Confidence score of detections.
                'labels' (N\ :sub:`det`,)   Predicted category number of detections in
                                            `torch.long` dtype.
                ======== ================== =================================================

                ========= ================= =================================================
                **y items**
                -----------------------------------------------------------------------------
                Key       Value shape        Description
                ========= ================= =================================================
                'bbox'    (N\ :sub:`gt`, 4) Bounding boxes of form (x1, y1, x2, y2)
                                            containing top left and bottom right coordinates.
                'labels'  (N\ :sub:`gt`,)   Category number of ground truths in `torch.long`
                                            dtype.
                'iscrowd' (N\ :sub:`gt`,)   Whether ground truth boxes are crowd ones or not.
                                            This key is optional.
                ========= ================= =================================================
        """
        self._check_matching_input(output)
        for pred, target in zip(*output):
            labels = target["labels"]
            gt_boxes = target["bbox"]
            gt_is_crowd = (
                target["iscrowd"].bool() if "iscrowd" in target else torch.zeros_like(labels, dtype=torch.bool)
            )
            gt_ignore = ~self._match_area_range(gt_boxes) | gt_is_crowd
            self._y_true_count += torch.bincount(labels[~gt_ignore], minlength=self._num_classes).to(
                device=self._device
            )

            # Matching logic of object detection mAP, according to COCO reference implementation.
            if len(pred["labels"]):
                best_detections_index = torch.argsort(pred["scores"], stable=True, descending=True)
                max_best_detections_index = torch.cat(
                    [
                        best_detections_index[pred["labels"][best_detections_index] == c][
                            : self._max_detections_per_image_per_class
                        ]
                        for c in range(self._num_classes)
                    ]
                )
                pred_boxes = pred["bbox"][max_best_detections_index]
                pred_labels = pred["labels"][max_best_detections_index]
                if not len(labels):
                    tp = torch.zeros(
                        (len(self._iou_thresholds), len(max_best_detections_index)),
                        dtype=torch.uint8,
                        device=self._device,
                    )
                    self._tps.append(tp)
                    self._fps.append(~tp & self._match_area_range(pred_boxes).to(self._device))
                else:
                    ious = self.box_iou(pred_boxes, gt_boxes, cast(torch.BoolTensor, gt_is_crowd))
                    category_no_match = labels.expand(len(pred_labels), -1) != pred_labels.view(-1, 1)
                    NO_MATCH = -3
                    ious[category_no_match] = NO_MATCH
                    ious = ious.unsqueeze(-1).repeat((1, 1, len(self._iou_thresholds)))
                    ious[ious < self._iou_thresholds] = NO_MATCH
                    IGNORANCE = -2
                    ious[:, gt_ignore] += IGNORANCE
                    for i in range(len(pred_labels)):
                        # Flip is done to give priority to the last item with maximal value,
                        # as torch.max selects the first one.
                        match_gts = ious[i].flip(0).max(0)
                        match_gts_indices = ious.size(1) - 1 - match_gts.indices
                        for t in range(len(self._iou_thresholds)):
                            if match_gts.values[t] > NO_MATCH and not gt_is_crowd[match_gts_indices[t]]:
                                ious[:, match_gts_indices[t], t] = NO_MATCH
                                ious[i, match_gts_indices[t], t] = match_gts.values[t]

                    max_ious = ious.max(1).values
                    self._tps.append((max_ious >= 0).T.to(dtype=torch.uint8, device=self._device))
                    self._fps.append(
                        ((max_ious <= NO_MATCH).T & self._match_area_range(pred_boxes)).to(
                            dtype=torch.uint8, device=self._device
                        )
                    )

                scores = pred["scores"][max_best_detections_index]
                if self._device.type == "mps" and scores.dtype == torch.double:
                    scores = scores.to(dtype=torch.float32)
                self._scores.append(scores.to(self._device))
                self._y_pred_labels.append(pred_labels.to(dtype=torch.int, device=self._device))

    @sync_all_reduce("_y_true_count")
    def _compute(self) -> torch.Tensor:
        pred_labels = _cat_and_agg_tensors(self._y_pred_labels, cast(Tuple[int], ()), torch.int, self._device)
        TP = _cat_and_agg_tensors(self._tps, (len(self._iou_thresholds),), torch.uint8, self._device)
        FP = _cat_and_agg_tensors(self._fps, (len(self._iou_thresholds),), torch.uint8, self._device)
        fp_precision = torch.double if self._device.type != "mps" else torch.float32
        scores = _cat_and_agg_tensors(self._scores, cast(Tuple[int], ()), fp_precision, self._device)

        average_precisions_recalls = -torch.ones(
            (2, self._num_classes, len(self._iou_thresholds)),
            device=self._device,
            dtype=fp_precision,
        )
        for cls in range(self._num_classes):
            if self._y_true_count[cls] == 0:
                continue

            cls_labels = pred_labels == cls
            if sum(cls_labels) == 0:
                average_precisions_recalls[:, cls] = 0.0
                continue

            recall, precision = self._compute_recall_and_precision(
                TP[..., cls_labels], FP[..., cls_labels], scores[cls_labels], self._y_true_count[cls]
            )
            average_precision_for_cls_per_iou_threshold = self._compute_average_precision(recall, precision)
            average_precisions_recalls[0, cls] = average_precision_for_cls_per_iou_threshold
            average_precisions_recalls[1, cls] = recall[..., -1]
        return average_precisions_recalls

    def compute(self) -> Tuple[float, float]:
        average_precisions_recalls = self._compute()
        if (average_precisions_recalls == -1).all():
            return -1.0, -1.0
        ap = average_precisions_recalls[0][average_precisions_recalls[0] > -1].mean().item()
        ar = average_precisions_recalls[1][average_precisions_recalls[1] > -1].mean().item()
        return ap, ar


class CommonObjectDetectionMetrics(MetricGroup):
    """
    Common Object Detection metrics. Included metrics are as follows:

    =============== ==========================================
    **Metric name**    **Description**
    =============== ==========================================
    AP@50..95       Average precision averaged over
                    .50 to.95 IOU thresholds
    AR-100          Average recall with maximum 100 detections
    AP@50           Average precision with IOU threshold=.50
    AP@75           Average precision with IOU threshold=.75
    AP-S            Average precision over small objects
                    (< 32px * 32px)
    AR-S            Average recall over small objects
    AP-M            Average precision over medium objects
                    (S < . < 96px * 96px)
    AR-M            Average recall over medium objects
    AP-L            Average precision over large objects
                    (M < . < 1e5px * 1e5px)
    AR-L            Average recall over large objects
                    greater than zero)
    AR-1            Average recall with maximum 1 detection
    AR-10           Average recall with maximum 10 detections
    =============== ==========================================

    .. versionadded:: 0.5.2
    """

    _state_dict_all_req_keys = ("metrics", "ap_50_95")

    ap_50_95: ObjectDetectionAvgPrecisionRecall

    def __init__(
        self,
        num_classes: int = 80,
        output_transform: Callable = lambda x: x,
        device: Union[str, torch.device] = torch.device("cpu"),
        skip_unrolling: bool = True,
    ):
        self.ap_50_95 = ObjectDetectionAvgPrecisionRecall(num_classes=num_classes, device=device)

        super().__init__(
            {
                "S": ObjectDetectionAvgPrecisionRecall(num_classes=num_classes, device=device, area_range="small"),
                "M": ObjectDetectionAvgPrecisionRecall(num_classes=num_classes, device=device, area_range="medium"),
                "L": ObjectDetectionAvgPrecisionRecall(num_classes=num_classes, device=device, area_range="large"),
                "1": ObjectDetectionAvgPrecisionRecall(
                    num_classes=num_classes, device=device, max_detections_per_image_per_class=1
                ),
                "10": ObjectDetectionAvgPrecisionRecall(
                    num_classes=num_classes, device=device, max_detections_per_image_per_class=10
                ),
            },
            output_transform,
            skip_unrolling=skip_unrolling,
        )

    def reset(self) -> None:
        super().reset()
        self.ap_50_95.reset()

    def update(self, output: Sequence[torch.Tensor]) -> None:
        super().update(output)
        self.ap_50_95.update(output)

    def compute(self) -> Dict[str, float]:
        average_precisions_recalls = self.ap_50_95._compute()

        average_precisions_50 = average_precisions_recalls[0, :, 0]
        average_precisions_75 = average_precisions_recalls[0, :, 5]
        if (average_precisions_50 == -1).all():
            AP_50 = AP_75 = AP_50_95 = AR_100 = -1.0
        else:
            AP_50 = average_precisions_50[average_precisions_50 > -1].mean().item()
            AP_75 = average_precisions_75[average_precisions_75 > -1].mean().item()
            AP_50_95 = average_precisions_recalls[0][average_precisions_recalls[0] > -1].mean().item()
            AR_100 = average_precisions_recalls[1][average_precisions_recalls[1] > -1].mean().item()

        result = super().compute()
        return {
            "AP@50..95": AP_50_95,
            "AR-100": AR_100,
            "AP@50": AP_50,
            "AP@75": AP_75,
            "AP-S": result["S"][0],
            "AR-S": result["S"][1],
            "AP-M": result["M"][0],
            "AR-M": result["M"][1],
            "AP-L": result["L"][0],
            "AR-L": result["L"][1],
            "AR-1": result["1"][1],
            "AR-10": result["10"][1],
        }



import os
import subprocess
import sys
from pathlib import Path

import pytest
import torch
from packaging.version import Version

import ignite.distributed as idist
from ignite.distributed.utils import has_hvd_support, has_native_dist_support, has_xla_support
from tests.ignite import is_mps_available_and_functional


def test_parallel_wrong_inputs():
    with pytest.raises(ValueError, match=r"Unknown backend 'abc'. Available backends:"):
        idist.Parallel(backend="abc")

    with pytest.raises(ValueError, match=r"If backend is None, argument 'nnodes' should be also None"):
        idist.Parallel(nnodes=2)

    with pytest.raises(ValueError, match=r"Argument nproc_per_node should positive"):
        idist.Parallel(backend="gloo", nproc_per_node=-1)

    with pytest.raises(ValueError, match=r"Argument nnodes should positive"):
        idist.Parallel(backend="gloo", nproc_per_node=1, nnodes=-1)

    with pytest.raises(ValueError, match=r"If number of nodes larger than one"):
        idist.Parallel(backend="gloo", nproc_per_node=1, nnodes=2)

    with pytest.raises(ValueError, match=r"Argument node_rank should be between 0 and"):
        idist.Parallel(backend="gloo", nproc_per_node=1, nnodes=2, node_rank=2)

    with pytest.raises(ValueError, match=r"If number of nodes larger than one, arguments master_addr and master_port"):
        idist.Parallel(backend="gloo", nproc_per_node=1, nnodes=2, node_rank=1)


@pytest.fixture()
def exec_filepath():
    fp = Path(__file__).parent / "check_idist_parallel.py"
    assert fp.exists()
    yield fp.as_posix()


def execute(cmd, env=None):
    import ignite

    env = dict(os.environ) if env is None else env
    env["PYTHONPATH"] = f"{os.path.dirname(ignite.__path__[0])}"
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)
    process.wait()
    if process.returncode != 0:
        print(str(process.stdout.read()) + str(process.stderr.read()))
        raise subprocess.CalledProcessError(returncode=process.returncode, cmd=cmd, stderr=process.stderr.read())
    return str(process.stdout.read()) + str(process.stderr.read())


@pytest.mark.skipif(not is_mps_available_and_functional(), reason="Skip if MPS not functional")
def test_check_idist_parallel_no_dist(exec_filepath):
    cmd = [sys.executable, "-u", exec_filepath]
    out = execute(cmd)
    assert "backend=None" in out
    assert "in 1 processes" in out
    assert "End of run" in out


def _test_check_idist_parallel_torch_launch(init_method, fp, backend, nprocs):
    # torchrun --nproc_per_node=nprocs tests/ignite/distributed/check_idist_parallel.py --backend=backend

    cmd = []
    if Version(torch.__version__) >= Version("1.10.0"):
        cmd += ["torchrun"]
    else:
        cmd += [
            sys.executable,
            "-m",
            "torch.distributed.launch",
            "--use_env",
        ]

    cmd += [
        f"--nproc_per_node={nprocs}",
        fp,
        f"--backend={backend}",
    ]

    if init_method is not None:
        cmd.append(f"--init_method={init_method}")

    out = execute(cmd)
    assert f"backend={backend}" in out
    assert f"in {nprocs} processes" in out
    assert "End of run" in out


@pytest.mark.distributed
@pytest.mark.skipif(not has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip because test uses torch launch")
@pytest.mark.parametrize("init_method", [None, "tcp://0.0.0.0:29500", "FILE"])
@pytest.mark.parametrize(
    "backend",
    ["gloo", pytest.param("nccl", marks=pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU"))],
)
def test_check_idist_parallel_torch_launch_n_procs_native(init_method, dirname, exec_filepath, backend):
    if init_method == "FILE":
        init_method = f"file://{dirname}/shared"

    np = torch.cuda.device_count() if torch.cuda.is_available() else 4
    _test_check_idist_parallel_torch_launch(init_method, exec_filepath, backend, np)


def _test_check_idist_parallel_hvdrun(fp, backend, nprocs):
    # horovodrun -np=nprocs python tests/ignite/distributed/check_idist_parallel.py --backend=backend

    cmd = [
        "horovodrun",
        "-np",
        f"{nprocs}",
        sys.executable,
        fp,
        f"--backend={backend}",
    ]

    out = execute(cmd)
    assert f"backend={backend}" in out
    assert f"in {nprocs} processes" in out
    assert "End of run" in out


@pytest.mark.distributed
@pytest.mark.skipif(not has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip because test uses horovodrun")
def test_check_idist_parallel_hvdrun_launch_n_procs(exec_filepath):
    np = 4 if not torch.cuda.is_available() else torch.cuda.device_count()
    _test_check_idist_parallel_hvdrun(exec_filepath, "horovod", np)


def _test_check_idist_parallel_spawn(fp, backend, nprocs):
    # python tests/ignite/distributed/check_idist_parallel.py --backend=backend --nproc_per_node=nprocs

    cmd = [sys.executable, fp, f"--backend={backend}", f"--nproc_per_node={nprocs}"]

    out = execute(cmd)
    assert f"backend={backend}" in out
    assert "Spawn function" in out
    assert f"in {nprocs} processes" in out
    if "xla" not in backend:
        assert "End of run" in out


@pytest.mark.distributed
@pytest.mark.skipif(not has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
@pytest.mark.parametrize(
    "backend",
    ["gloo", pytest.param("nccl", marks=pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU"))],
)
def test_check_idist_parallel_spawn_n_procs_native(exec_filepath, backend):
    np = 4 if not torch.cuda.is_available() else torch.cuda.device_count()
    _test_check_idist_parallel_spawn(exec_filepath, backend, np)


@pytest.mark.distributed
@pytest.mark.skipif(not has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_smoke_test_check_idist_parallel_spawn_multinode_n_procs_gloo(exec_filepath):
    # Just a smoke test from check_idist_parallel.py for an emulated multi-node configuration
    cmd1 = "export CUDA_VISIBLE_DEVICES= && "
    cmd1 += f'bash -c "{sys.executable} {exec_filepath} --backend=gloo --nproc_per_node=2 '
    cmd1 += '--nnodes=2 --node_rank=0 --master_addr=localhost --master_port=3344 &"'
    os.system(cmd1)

    cmd2 = [
        sys.executable,
        exec_filepath,
        "--backend=gloo",
        "--nproc_per_node=2",
        "--nnodes=2",
        "--node_rank=1",
        "--master_addr=localhost",
        "--master_port=3344",
    ]
    env = dict(os.environ)
    env["CUDA_VISIBLE_DEVICES"] = ""
    out = execute(cmd2, env=env)

    assert "backend=gloo" in out
    assert "nproc_per_node: 2" in out
    assert "nnodes: 2" in out
    assert "master_addr: localhost" in out
    assert "master_port: 3344" in out
    assert "End of run" in out


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not has_xla_support, reason="Skip if no PyTorch XLA package")
def test_check_idist_parallel_spawn_n_procs_xla(exec_filepath):
    n = int(os.environ["NUM_TPU_WORKERS"])
    if n > 1:
        _test_check_idist_parallel_spawn(exec_filepath, "xla-tpu", n)


@pytest.mark.distributed
@pytest.mark.skipif(not has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_check_idist_parallel_spawn_n_procs_hvd(exec_filepath):
    np = 4 if not torch.cuda.is_available() else torch.cuda.device_count()
    _test_check_idist_parallel_spawn(exec_filepath, "horovod", np)


def _test_func(index, ws, device, backend, true_init_method):
    assert 0 <= index < ws
    assert index == idist.get_local_rank()
    assert ws == idist.get_world_size()
    assert torch.device(device).type == idist.device().type
    assert backend == idist.backend()

    if idist.model_name() == "native-dist":
        from ignite.distributed.utils import _model

        assert _model._init_method == true_init_method


@pytest.mark.distributed
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
@pytest.mark.skipif(not has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.parametrize("init_method", ["env://", "tcp://0.0.0.0:29500", "FILE"])
@pytest.mark.parametrize(
    "backend",
    ["gloo", pytest.param("nccl", marks=pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU"))],
)
def test_idist_parallel_spawn_n_procs_native(init_method, backend, dirname):
    if init_method == "FILE":
        init_method = f"file://{dirname}/shared"

    nproc_per_node = torch.cuda.device_count() if torch.cuda.is_available() else 4
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    with idist.Parallel(backend=backend, nproc_per_node=nproc_per_node, init_method=init_method) as parallel:
        parallel.run(_test_func, ws=nproc_per_node, device=device, backend=backend, true_init_method=init_method)


@pytest.mark.distributed
@pytest.mark.skipif("WORLD_SIZE" not in os.environ, reason="Skip if not launched as multiproc")
@pytest.mark.skipif(not has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.parametrize("init_method", ["env://", "tcp://0.0.0.0:29500", "FILE"])
@pytest.mark.parametrize(
    "backend",
    ["gloo", pytest.param("nccl", marks=pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU"))],
)
def test_idist_parallel_n_procs_native(init_method, backend, get_fixed_dirname, local_rank, world_size):
    if init_method == "FILE":
        init_method = f"file://{get_fixed_dirname('idist_parallel_n_procs_native')}/shared"

    os.environ["RANK"] = str(local_rank)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    with idist.Parallel(backend=backend, init_method=init_method) as parallel:
        parallel.run(_test_func, ws=world_size, device=device, backend=backend, true_init_method=init_method)


@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_idist_parallel_no_dist():
    device = idist.device()
    with idist.Parallel(backend=None) as parallel:
        parallel.run(_test_func, ws=1, device=device, backend=None, true_init_method=None)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not has_xla_support, reason="Skip if no PyTorch XLA package")
def test_idist_parallel_spawn_params_xla():
    res = idist.Parallel._setup_spawn_params(
        nproc_per_node=8, nnodes=None, node_rank=None, master_addr=None, master_port=None, start_method="fork"
    )
    assert "nproc_per_node" in res and res["nproc_per_node"] == 8
    assert "start_method" in res and res["start_method"] == "fork"

    with idist.Parallel(backend="xla-tpu", nproc_per_node=8, start_method="fork") as parallel:
        assert parallel.backend == "xla-tpu"
        res = parallel._spawn_params
        assert "nproc_per_node" in res and res["nproc_per_node"] == 8
        assert "start_method" in res and res["start_method"] == "fork"



import sys
import time

import pytest

from ignite.engine import Engine, Events
from ignite.handlers import Timer

if sys.platform.startswith("darwin"):
    pytest.skip("Skip if on MacOS", allow_module_level=True)


def test_timer():
    sleep_t = 0.2
    n_iter = 3

    def _train_func(engine, batch):
        time.sleep(sleep_t)

    def _test_func(engine, batch):
        time.sleep(sleep_t)

    trainer = Engine(_train_func)
    tester = Engine(_test_func)

    t_total = Timer()
    t_batch = Timer(average=True)
    t_train = Timer()

    t_total.attach(trainer)
    t_batch.attach(
        trainer, pause=Events.ITERATION_COMPLETED, resume=Events.ITERATION_STARTED, step=Events.ITERATION_COMPLETED
    )
    t_train.attach(trainer, pause=Events.EPOCH_COMPLETED, resume=Events.EPOCH_STARTED)

    @trainer.on(Events.EPOCH_COMPLETED)
    def run_validation(trainer):
        tester.run(range(n_iter))

    # Run "training"
    trainer.run(range(n_iter))

    assert pytest.approx(t_total.value(), abs=1e-1) == 2 * n_iter * sleep_t
    assert pytest.approx(t_batch.value(), abs=1e-1) == sleep_t
    assert pytest.approx(t_train.value(), abs=1e-1) == n_iter * sleep_t

    t_total.reset()
    assert pytest.approx(t_total.value(), abs=1e-1) == 0.0



import torch

from ignite.distributed.comp_models.base import _torch_version_gt_112


def is_mps_available_and_functional():
    if not _torch_version_gt_112:
        return False

    if not torch.backends.mps.is_available():
        return False
    try:
        # Try to allocate a small tensor on the MPS device
        torch.tensor([1.0], device="mps")
        return True
    except RuntimeError:
        return False



import os

import pytest
import torch
from sklearn.metrics import DistanceMetric

import ignite.distributed as idist
from ignite.engine import Engine
from ignite.metrics.regression import ManhattanDistance


def test_wrong_input_shapes():
    m = ManhattanDistance()

    with pytest.raises(ValueError, match=r"Input data shapes should be the same, but given"):
        m.update((torch.rand(4), torch.rand(4, 1)))

    with pytest.raises(ValueError, match=r"Input data shapes should be the same, but given"):
        m.update((torch.rand(4, 1), torch.rand(4)))


def test_manhattan_distance(available_device):
    preds = [torch.randn(4) for _ in range(4)]
    ground_truth = torch.randn(4)

    m = ManhattanDistance(device=available_device)
    assert m._device == torch.device(available_device)

    from sklearn.metrics import DistanceMetric

    manhattan = DistanceMetric.get_metric("manhattan")

    total_sum = 0.0
    v1 = []
    v2 = []

    for pred in preds:
        m.update((pred, ground_truth))
        error = torch.abs(pred - ground_truth).sum()
        total_sum += error
        assert m.compute() == pytest.approx(total_sum)

        v1.extend(pred.cpu().numpy())
        v2.extend(ground_truth.cpu().numpy())
        assert manhattan.pairwise([v1, v2])[0][1] == pytest.approx(total_sum)


@pytest.mark.parametrize("n_times", range(3))
@pytest.mark.parametrize(
    "test_case",
    [
        (torch.rand(size=(100,)), torch.rand(size=(100,)), 10),
        (torch.rand(size=(100, 1)), torch.rand(size=(100, 1)), 20),
    ],
)
def test_integration_manhattan_distance(test_case, n_times, available_device):
    y_pred, y, batch_size = test_case

    def update_fn(engine, batch):
        idx = (engine.state.iteration - 1) * batch_size
        return y_pred[idx : idx + batch_size], y[idx : idx + batch_size]

    engine = Engine(update_fn)

    metric = ManhattanDistance(device=available_device)
    assert metric._device == torch.device(available_device)
    metric.attach(engine, "md")

    data = list(range(y_pred.shape[0] // batch_size))
    result = engine.run(data, max_epochs=1).metrics["md"]

    flat_pred = y_pred.view(-1).cpu()
    flat_y = y.view(-1).cpu()
    expected = torch.sum(torch.abs(flat_pred - flat_y)).item()

    assert expected == pytest.approx(result)


def test_error_is_not_nan(available_device):
    m = ManhattanDistance(device=available_device)
    assert m._device == torch.device(available_device)
    m.update((torch.zeros(4), torch.zeros(4)))
    assert not (torch.isnan(m._sum_of_errors).any() or torch.isinf(m._sum_of_errors).any()), m._sum_of_errors


def _test_distrib_compute(device):
    rank = idist.get_rank()

    manhattan = DistanceMetric.get_metric("manhattan")

    def _test(metric_device):
        metric_device = torch.device(metric_device)
        m = ManhattanDistance(device=metric_device)

        y_pred = torch.randint(0, 10, size=(10,), device=device).float()
        y = torch.randint(0, 10, size=(10,), device=device).float()

        m.update((y_pred, y))

        # gather y_pred, y
        y_pred = idist.all_gather(y_pred)
        y = idist.all_gather(y)

        np_y_pred = y_pred.cpu().numpy()
        np_y = y.cpu().numpy()
        res = m.compute()
        assert manhattan.pairwise([np_y_pred, np_y])[0][1] == pytest.approx(res)

    for i in range(3):
        torch.manual_seed(10 + rank + i)
        _test("cpu")
        if device.type != "xla":
            _test(idist.device())


def _test_distrib_integration(device):
    rank = idist.get_rank()

    manhattan = DistanceMetric.get_metric("manhattan")

    def _test(n_epochs, metric_device):
        metric_device = torch.device(metric_device)
        n_iters = 80
        batch_size = 16

        y_true = torch.rand(size=(n_iters * batch_size,)).to(device)
        y_preds = torch.rand(size=(n_iters * batch_size,)).to(device)

        def update(engine, i):
            return (
                y_preds[i * batch_size : (i + 1) * batch_size],
                y_true[i * batch_size : (i + 1) * batch_size],
            )

        engine = Engine(update)

        m = ManhattanDistance(device=metric_device)
        m.attach(engine, "md")

        data = list(range(n_iters))
        engine.run(data=data, max_epochs=n_epochs)

        y_preds = idist.all_gather(y_preds)
        y_true = idist.all_gather(y_true)

        assert "md" in engine.state.metrics

        res = engine.state.metrics["md"]
        if isinstance(res, torch.Tensor):
            res = res.cpu().numpy()

        np_y_true = y_true.cpu().numpy()
        np_y_preds = y_preds.cpu().numpy()

        assert pytest.approx(res) == manhattan.pairwise([np_y_preds, np_y_true])[0][1]

    metric_devices = ["cpu"]
    if device.type != "xla":
        metric_devices.append(idist.device())
    for metric_device in metric_devices:
        for i in range(2):
            torch.manual_seed(12 + rank + i)
            _test(n_epochs=1, metric_device=metric_device)
            _test(n_epochs=2, metric_device=metric_device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor):
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_distrib_compute, (device,), np=nproc, do_init=True)
    gloo_hvd_executor(_test_distrib_integration, (device,), np=nproc, do_init=True)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_gloo_cpu_or_gpu(distributed_context_multi_node_gloo):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_nccl_gpu(distributed_context_multi_node_nccl):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_single_device_xla():
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


def _test_distrib_xla_nprocs(index):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



import argparse

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
from torch.optim.lr_scheduler import StepLR
from torch.utils.data import DataLoader, Dataset
from torchvision import datasets

from ignite.engine import Engine, Events

from ignite.handlers import ProgressBar
from ignite.handlers.param_scheduler import LRScheduler
from ignite.metrics import Accuracy, RunningAverage
from ignite.utils import manual_seed


class SiameseNetwork(nn.Module):
    # update Siamese Network implementation in accordance with the dataset
    """
    Siamese network for image similarity estimation.
    The network is composed of two identical networks, one for each input.
    The output of each network is concatenated and passed to a linear layer.
    The output of the linear layer passed through a sigmoid function.
    `"FaceNet" <https://arxiv.org/pdf/1503.03832.pdf>`_ is a variant of the Siamese network.
    This implementation varies from FaceNet as we use the `ResNet-18` model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`
    as our feature extractor.
    In addition we use CIFAR10 dataset along with TripletMarginLoss
    """

    def __init__(self):
        super(SiameseNetwork, self).__init__()
        # get resnet model
        self.resnet = torchvision.models.resnet34(weights=None)
        fc_in_features = self.resnet.fc.in_features

        # changing the FC layer of resnet model to a linear layer
        self.resnet.fc = nn.Identity()

        # add linear layers to compare between the features of the two images
        self.fc = nn.Sequential(
            nn.Linear(fc_in_features, 256),
            nn.ReLU(inplace=True),
            nn.Linear(256, 10),
            nn.ReLU(inplace=True),
        )

        # initialise relu activation
        self.relu = nn.ReLU()

        # initialize the weights
        self.resnet.apply(self.init_weights)
        self.fc.apply(self.init_weights)

    def init_weights(self, m):
        if isinstance(m, nn.Linear):
            nn.init.xavier_uniform_(m.weight)
            m.bias.data.fill_(0.01)

    def forward_once(self, x):
        output = self.resnet(x)
        output = output.view(output.size()[0], -1)
        return output

    def forward(self, input1, input2, input3):
        # pass the input through resnet
        output1 = self.forward_once(input1)
        output2 = self.forward_once(input2)
        output3 = self.forward_once(input3)

        # pass the output of resnet to sigmoid layer
        output1 = self.fc(output1)
        output2 = self.fc(output2)
        output3 = self.fc(output3)

        return output1, output2, output3


class MatcherDataset(Dataset):
    # following class implements data downloading and handles preprocessing
    def __init__(self, root, train, download=False):
        super(MatcherDataset, self).__init__()

        # get CIFAR10 dataset
        self.dataset = datasets.CIFAR10(root, train=train, download=download)

        # convert data from numpy array to Tensor
        self.data = torch.from_numpy(self.dataset.data)

        # shift the dimensions of dataset to match the initial input layer dimensions
        self.data = torch.movedim(self.data, (0, 1, 2, 3), (0, 2, 3, 1))

        # convert targets list to torch Tensor
        self.dataset.targets = torch.tensor(self.dataset.targets)

        self.group_examples()

    def group_examples(self):
        """
        To ease the accessibility of data based on the class, we will use `group_examples` to group
        examples based on class. The data classes have already been mapped to numeric values and
        so are the target outputs for each training input

        Every key in `grouped_examples` corresponds to a class in CIFAR10 dataset. For every key in
        `grouped_examples`, every value will conform to all of the indices for the CIFAR10
        dataset examples that correspond to that key.
        """

        # get the targets from CIFAR10 dataset
        np_arr = np.array(self.dataset.targets)

        # group examples based on class
        self.grouped_examples = {}
        for i in range(0, 10):
            self.grouped_examples[i] = np.where((np_arr == i))[0]

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, index):
        """
        For every sample in the batch we select 3 images. First one is the anchor image
        which is the image obtained from the current index. We also obtain the label of
        anchor image.

        Now we select two random images, one belonging to the same class as that of the
        anchor image (named as positive_image) and the other belonging to a different class
        than that of the anchor image (named as negative_image). We return the anchor image,
        positive image, negative image and anchor label.
        """

        # obtain the anchor image
        anchor_image = self.data[index].float()

        # obtain the class label of the anchor image
        anchor_label = self.dataset.targets[index]
        anchor_label = int(anchor_label.item())

        # find a label which is different from anchor_label
        labels = list(range(0, 10))
        labels.remove(anchor_label)
        neg_index = torch.randint(0, 9, (1,)).item()
        neg_label = labels[neg_index]

        # get a random index from the range range of indices
        random_index = torch.randint(0, len(self.grouped_examples[anchor_label]), (1,)).item()

        # get the index of image in actual data using the anchor label and random index
        positive_index = self.grouped_examples[anchor_label][random_index]

        # choosing a random image using positive_index
        positive_image = self.data[positive_index].float()

        # get a random index from the range range of indices
        random_index = torch.randint(0, len(self.grouped_examples[neg_label]), (1,)).item()

        # get the index of image in actual data using the negative label and random index
        negative_index = self.grouped_examples[neg_label][random_index]

        # choosing a random image using negative_index
        negative_image = self.data[negative_index].float()

        return anchor_image, positive_image, negative_image, anchor_label


def pairwise_distance(input1, input2):
    dist = input1 - input2
    dist = torch.pow(dist, 2)
    return dist


def calculate_loss(input1, input2):
    output = pairwise_distance(input1, input2)
    loss = torch.sum(output, 1)
    loss = torch.sqrt(loss)
    return loss


def run(args, model, device, optimizer, train_loader, test_loader, lr_scheduler):
    # using Triplet Margin Loss
    criterion = nn.TripletMarginLoss(p=2, margin=2.8)

    # define model training step
    def train_step(engine, batch):
        model.train()
        anchor_image, positive_image, negative_image, anchor_label = batch
        anchor_image = anchor_image.to(device)
        positive_image, negative_image = positive_image.to(device), negative_image.to(device)
        anchor_label = anchor_label.to(device)
        optimizer.zero_grad()
        anchor_out, positive_out, negative_out = model(anchor_image, positive_image, negative_image)
        loss = criterion(anchor_out, positive_out, negative_out)
        loss.backward()
        optimizer.step()
        return loss

    # define model testing step
    def test_step(engine, batch):
        model.eval()
        with torch.no_grad():
            anchor_image, _, _, anchor_label = batch
            anchor_image = anchor_image.to(device)
            anchor_label = anchor_label.to(device)
            other_image = []
            other_label = []
            y_true = []
            for i in range(anchor_image.shape[0]):
                index = torch.randint(0, anchor_image.shape[0], (1,)).item()
                img = anchor_image[index]
                label = anchor_label[index]
                other_image.append(img)
                other_label.append(label)
                if anchor_label[i] == other_label[i]:
                    y_true.append(1)
                else:
                    y_true.append(0)
            other = torch.stack(other_image)
            other_label = torch.tensor(other_label)
            other, other_label = other.to(device), other_label.to(device)
            anchor_out, other_out, _ = model(anchor_image, other, other)
            test_loss = calculate_loss(anchor_out, other_out)
            y_pred = torch.where(test_loss < 3, 1, 0)
            y_true = torch.tensor(y_true)
            return [y_pred, y_true]

    # create engines for trainer and evaluator
    trainer = Engine(train_step)
    evaluator = Engine(test_step)

    # attach Running Average Loss metric to trainer and evaluator engines
    RunningAverage(output_transform=lambda x: x).attach(trainer, "loss")
    Accuracy(output_transform=lambda x: x).attach(evaluator, "accuracy")

    # attach progress bar to trainer with loss
    pbar1 = ProgressBar()
    pbar1.attach(trainer, metric_names=["loss"])

    # attach progress bar to evaluator
    pbar2 = ProgressBar()
    pbar2.attach(evaluator)

    # attach LR Scheduler to trainer engine
    trainer.add_event_handler(Events.ITERATION_STARTED, lr_scheduler)

    # event handler triggers evauator at end of every epoch
    @trainer.on(Events.EPOCH_COMPLETED(every=args.log_interval))
    def test(engine):
        state = evaluator.run(test_loader)
        print(f'Test Accuracy: {state.metrics["accuracy"]}')

    # run the trainer
    trainer.run(train_loader, max_epochs=args.epochs)


def main():
    # adds training defaults and support for terminal arguments
    parser = argparse.ArgumentParser(description="PyTorch Siamese network Example")
    parser.add_argument(
        "--batch-size", type=int, default=256, metavar="N", help="input batch size for training (default: 64)"
    )
    parser.add_argument(
        "--test-batch-size", type=int, default=256, metavar="N", help="input batch size for testing (default: 1000)"
    )
    parser.add_argument("--epochs", type=int, default=10, metavar="N", help="number of epochs to train (default: 14)")
    parser.add_argument("--lr", type=float, default=1.0, metavar="LR", help="learning rate (default: 1.0)")
    parser.add_argument(
        "--gamma", type=float, default=0.95, metavar="M", help="Learning rate step gamma (default: 0.7)"
    )
    parser.add_argument("--no-cuda", action="store_true", default=False, help="disables CUDA training")
    parser.add_argument("--no-mps", action="store_true", default=False, help="disables macOS GPU training")
    parser.add_argument("--dry-run", action="store_true", default=False, help="quickly check a single pass")
    parser.add_argument("--seed", type=int, default=1, metavar="S", help="random seed (default: 1)")
    parser.add_argument(
        "--log-interval",
        type=int,
        default=1,
        metavar="N",
        help="how many batches to wait before logging training status",
    )
    parser.add_argument("--save-model", action="store_true", default=False, help="For Saving the current Model")
    parser.add_argument("--num-workers", default=4, help="number of processes generating parallel batches")
    args = parser.parse_args()

    # set manual seed
    manual_seed(args.seed)

    # set device
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

    # data loading
    train_dataset = MatcherDataset("../data", train=True, download=True)
    test_dataset = MatcherDataset("../data", train=False)
    train_loader = DataLoader(train_dataset, shuffle=True, batch_size=args.batch_size, num_workers=args.num_workers)
    test_loader = DataLoader(test_dataset, batch_size=args.test_batch_size, num_workers=args.num_workers)

    # set model parameters
    model = SiameseNetwork().to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=args.lr)
    scheduler = StepLR(optimizer, step_size=15, gamma=args.gamma)
    lr_scheduler = LRScheduler(scheduler)

    # call run function
    run(args, model, device, optimizer, train_loader, test_loader, lr_scheduler)


if __name__ == "__main__":
    main()



#
# Tests :
# For all images
#     can import torch and its version == required one
#     can import ignite and its version == required one
# for all -vision images
#     can import opencv without driver issue
# for all horovod images
#     can import horovod and its version == required one
#
import argparse
import importlib
import os


def check_package(package_name, expected_version=None):
    mod = importlib.import_module(package_name)

    if expected_version is not None:
        assert hasattr(mod, "__version__"), f"Imported package {package_name} does not have __version__ attribute"
        version = mod.__version__
        # Remove all +something from the version name: e.g torch 2.5.1+cu124
        if "+" in version:
            old_version = version
            version = version.split("+")[0]
            print(f"Transformed version: {old_version} -> {version}")
        assert (
            version == expected_version
        ), f"Version mismatch for package {package_name}: got {version} but expected {expected_version}"


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Check docker image script")
    parser.add_argument("image", type=str, help="Docker image to check")
    args = parser.parse_args()

    docker_image_name = args.image
    name, version = docker_image_name.split(":")
    assert version != "latest", version
    torch_version, ignite_version = version.split("-")
    _, image_type = name.split("/")

    check_package("torch", expected_version=torch_version)
    check_package("ignite", expected_version=ignite_version)

    if "hvd" in image_type:
        assert "HVD_VERSION" in os.environ
        val = os.environ["HVD_VERSION"]
        hvd_version = val if val[0] != "v" else val[1:]
        check_package("horovod", expected_version=hvd_version)

    if "msdp" in image_type:
        assert "MSDP_VERSION" in os.environ
        val = os.environ["MSDP_VERSION"]
        hvd_version = val if val[0] != "v" else val[1:]
        check_package("deepspeed", expected_version=hvd_version)

    if "vision" in image_type:
        check_package("cv2")

    if "nlp" in image_type:
        check_package("transformers")

    if "apex" in image_type:
        check_package("apex")



import os
from pathlib import Path

import brevitas.nn as qnn
import torch
import torch.nn as nn
from pact import PACTReLU
from torchvision import datasets, models
from torchvision.transforms import Compose, Normalize, Pad, RandomCrop, RandomHorizontalFlip, ToTensor

train_transform = Compose(
    [
        Pad(4),
        RandomCrop(32, fill=128),
        RandomHorizontalFlip(),
        ToTensor(),
        Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
    ]
)

test_transform = Compose([ToTensor(), Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225))])


def get_train_test_datasets(path):
    path = Path(path)
    if not path.exists():
        path.mkdir(parents=True)
        download = True
    else:
        download = True if len(os.listdir(path)) < 1 else False

    train_ds = datasets.CIFAR10(root=path, train=True, download=download, transform=train_transform)
    test_ds = datasets.CIFAR10(root=path, train=False, download=False, transform=test_transform)

    return train_ds, test_ds


def get_model(name):
    __dict__ = globals()

    if name in models.__dict__:
        fn = models.__dict__[name]
    elif name in ["resnet18_QAT_8b", "resnet18_QAT_6b", "resnet18_QAT_5b", "resnet18_QAT_4b"]:
        fn = __dict__[name]
    else:
        raise RuntimeError("Unknown model name {}".format(name))

    return fn(num_classes=10)


# Below code is taken from https://discuss.pytorch.org/t/evaluator-returns-nan/107972/3


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1, weight_bit_width=8):
    """3x3 convolution with padding"""
    return qnn.QuantConv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
        weight_bit_width=weight_bit_width,
    )


def conv1x1(in_planes, out_planes, stride=1, weight_bit_width=8):
    """1x1 convolution"""
    return qnn.QuantConv2d(
        in_planes, out_planes, kernel_size=1, stride=stride, bias=False, weight_bit_width=weight_bit_width
    )


def make_PACT_relu(bit_width=8):
    relu = qnn.QuantReLU(bit_width=bit_width)
    relu.act_impl = PACTReLU()
    return relu


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
        bit_width=8,
    ):
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride, weight_bit_width=bit_width)
        self.bn1 = norm_layer(planes)
        self.relu = make_PACT_relu(bit_width=bit_width)
        self.conv2 = conv3x3(planes, planes, weight_bit_width=bit_width)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion = 4

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
        bit_width=8,
    ):
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width, weight_bit_width=bit_width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation, weight_bit_width=bit_width)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion, weight_bit_width=bit_width)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = make_PACT_relu(bit_width=bit_width)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet_QAT_Xb(nn.Module):
    def __init__(
        self,
        block,
        layers,
        num_classes=1000,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        norm_layer=None,
        bit_width=8,
    ):
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                "or a 3-element tuple, got {}".format(replace_stride_with_dilation)
            )
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = qnn.QuantConv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = make_PACT_relu()
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0], bit_width=bit_width)
        self.layer2 = self._make_layer(
            block, 128, layers[1], stride=2, dilate=replace_stride_with_dilation[0], bit_width=bit_width
        )
        self.layer3 = self._make_layer(
            block, 256, layers[2], stride=2, dilate=replace_stride_with_dilation[1], bit_width=bit_width
        )
        self.layer4 = self._make_layer(
            block, 512, layers[3], stride=2, dilate=replace_stride_with_dilation[2], bit_width=bit_width
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):  # qnn.QuantConv2d includes nn.Conv2d inside.
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False, bit_width=8):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride, weight_bit_width=bit_width),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
                bit_width=bit_width,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                    bit_width=bit_width,
                )
            )

        return nn.Sequential(*layers)

    def _forward_impl(self, x):
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


def _resnet_QAT_Xb(block, layers, **kwargs):
    model = ResNet_QAT_Xb(block, layers, **kwargs)
    return model


def resnet18_QAT_8b(*args, **kwargs):
    return _resnet_QAT_Xb(BasicBlock, [2, 2, 2, 2], **kwargs)


def resnet18_QAT_6b(*args, **kwargs):
    return _resnet_QAT_Xb(BasicBlock, [2, 2, 2, 2], bit_width=6, **kwargs)


def resnet18_QAT_5b(*args, **kwargs):
    return _resnet_QAT_Xb(BasicBlock, [2, 2, 2, 2], bit_width=5, **kwargs)


def resnet18_QAT_4b(*args, **kwargs):
    return _resnet_QAT_Xb(BasicBlock, [2, 2, 2, 2], bit_width=4, **kwargs)



import os

import pytest
import torch

from ignite.distributed.comp_models import has_xla_support

if not has_xla_support:
    pytest.skip("Skip if no XLA support", allow_module_level=True)
else:
    from ignite.distributed.comp_models.xla import _XlaDistModel


@pytest.mark.tpu
@pytest.mark.skipif(not has_xla_support, reason="Skip if no PyTorch XLA package")
def test__xla_model():
    available_backends = _XlaDistModel.available_backends
    assert "xla-tpu" in available_backends

    with pytest.raises(ValueError, match=r"Backend should be one of"):
        _XlaDistModel.create_from_backend("abc")


def _test_xla_spawn_fn(local_rank, world_size, device):
    from ignite.distributed.utils import _model

    assert isinstance(_model, _XlaDistModel), f"{type(_model)} vs _XlaDistModel"

    assert _model.get_local_rank() == local_rank
    assert _model.get_world_size() == world_size
    d = _model.device()
    assert isinstance(d, torch.device) and d.type == device

    assert _model.get_rank() == local_rank
    assert _model.get_nproc_per_node() == world_size
    assert _model.get_node_rank() == 0
    assert _model.get_nnodes() == 1


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not has_xla_support, reason="Skip if no PyTorch XLA package")
def test__xla_dist_model_spawn_one_proc():
    try:
        _XlaDistModel.spawn(_test_xla_spawn_fn, args=(1, "xla"), kwargs_dict={}, nproc_per_node=1)
    except SystemExit:
        pass


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not has_xla_support, reason="Skip if no PyTorch XLA package")
def test__xla_dist_model_spawn_n_procs():
    n = int(os.environ["NUM_TPU_WORKERS"])
    try:
        _XlaDistModel.spawn(_test_xla_spawn_fn, args=(n, "xla"), kwargs_dict={}, nproc_per_node=n)
    except SystemExit:
        pass


def _assert_model(model, true_conf):
    assert model.device() == true_conf["device"]
    assert model.get_local_rank() == true_conf["local_rank"]
    assert model.get_rank() == true_conf["rank"]
    assert model.get_world_size() == true_conf["world_size"]

    assert model.get_node_rank() == true_conf["node_index"]
    assert model.get_nnodes() == true_conf["nnodes"]
    assert model.get_nproc_per_node() == true_conf["nproc_per_node"]


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not has_xla_support, reason="Skip if no PyTorch XLA package")
def test__xla_dist_model_create_from_backend():
    # without spawn
    model = _XlaDistModel.create_from_backend("xla-tpu")

    import torch_xla.core.xla_model as xm

    _assert_model(
        model,
        {
            "device": xm.xla_device(),
            "local_rank": 0,
            "rank": 0,
            "world_size": 1,
            "node_index": 0,
            "nnodes": 1,
            "nproc_per_node": 1,
        },
    )

    model.finalize()


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not has_xla_support, reason="Skip if no PyTorch XLA package")
def test__xla_dist_model_create_from_context():
    # without spawn
    model = _XlaDistModel.create_from_context()

    assert model.backend() == "xla-tpu"

    import torch_xla.core.xla_model as xm

    _assert_model(
        model,
        {
            "device": xm.xla_device(),
            "local_rank": 0,
            "rank": 0,
            "world_size": 1,
            "node_index": 0,
            "nnodes": 1,
            "nproc_per_node": 1,
        },
    )


def _test__xla_dist_model_create_from_context_in_child_proc(index):
    model = _XlaDistModel.create_from_context()

    assert model.backend() == "xla-tpu"

    import torch_xla.core.xla_model as xm

    _assert_model(
        model,
        {
            "device": xm.xla_device(),
            "local_rank": index,
            "rank": xm.get_ordinal(),
            "world_size": xm.xrt_world_size(),
            "node_index": 0,
            "nnodes": 1,
            "nproc_per_node": xm.xrt_world_size(),
        },
    )


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not has_xla_support, reason="Skip if no PyTorch XLA package")
def test__xla_dist_model_create_from_context_in_child_proc(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test__xla_dist_model_create_from_context_in_child_proc, args=(), nprocs=n)


def main_fold(fold):
    import time

    import torch.nn as nn
    import torch.optim as optim
    import torch_xla.core.xla_model as xm

    from ignite.engine import Engine

    device = xm.xla_device(fold)

    comp_model = _XlaDistModel.create_from_context()
    assert comp_model.device() == device

    model = nn.Linear(100, 10)

    model.to(device)  # Move model before creating optimizer
    optimizer = optim.SGD(model.parameters(), lr=0.01, momentum=0.9)

    def training_step(engine, _):
        data = torch.rand(4, 100, device=device)
        model.train()
        data = data.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = output.sum()
        loss.backward()
        xm.optimizer_step(optimizer, barrier=True)
        return loss.item()

    trainer = Engine(training_step)

    # THIS CAN BE A CAUSE OF CRASH if DEVICE is OTHER THAN device
    tensor = torch.tensor([fold + 1.0], dtype=torch.float).to(comp_model.device())
    xm.all_reduce("max", [tensor])

    time.sleep(0.01 * fold)

    trainer.run([0] * 100, max_epochs=2)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not has_xla_support, reason="Skip if no PyTorch XLA package")
def test__xla_dist_model_run_parallel_n_threads_without_sync():
    # tests issue : https://github.com/pytorch/ignite/issues/1096
    import torch_xla.core.xla_model as xm
    from joblib import delayed, Parallel

    devices = xm.get_xla_supported_devices()
    folds = 1
    d = 0
    if len(devices) > 5:
        folds = 5
        d = 1
    Parallel(n_jobs=folds, backend="threading")(delayed(main_fold)(i + d) for i in range(folds))



import math
import os
from collections import defaultdict
from unittest.mock import ANY, call, MagicMock, patch

import clearml
import pytest
import torch
from clearml.binding.frameworks import WeightsFileHandler
from clearml.model import Framework

import ignite.distributed as idist

from ignite.engine import Engine, Events, State
from ignite.handlers import Checkpoint
from ignite.handlers.clearml_logger import (
    ClearMLLogger,
    ClearMLSaver,
    global_step_from_engine,
    GradsHistHandler,
    GradsScalarHandler,
    OptimizerParamsHandler,
    OutputHandler,
    WeightsHistHandler,
    WeightsScalarHandler,
)


def test_no_clearml():
    with patch.dict("sys.modules", {"clearml": None, "trains": None}):
        with pytest.raises(ModuleNotFoundError, match=r"This contrib module requires clearml to be installed."):
            ClearMLSaver()

        with pytest.raises(ModuleNotFoundError, match=r"This contrib module requires clearml to be installed."):
            ClearMLLogger()

    with patch.dict("sys.modules", {"clearml.binding.frameworks.tensorflow_bind": None}):
        with pytest.raises(ModuleNotFoundError, match=r"This contrib module requires clearml to be installed."):
            ClearMLLogger()

    with patch.dict("sys.modules", {"clearml.binding.frameworks": None, "trains.binding.frameworks": None}):
        with pytest.raises(ModuleNotFoundError, match=r"This contrib module requires clearml to be installed."):
            ClearMLSaver.__call__(None, {}, "")


def test_optimizer_params_handler_wrong_setup():
    with pytest.raises(TypeError):
        OptimizerParamsHandler(optimizer=None)

    optimizer = MagicMock(spec=torch.optim.Optimizer)
    handler = OptimizerParamsHandler(optimizer=optimizer)

    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler OptimizerParamsHandler works only with ClearMLLogger"):
        handler(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_optimizer_params():
    optimizer = torch.optim.SGD([torch.tensor(0.0)], lr=0.01)
    wrapper = OptimizerParamsHandler(optimizer=optimizer, param_name="lr")
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()
    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.iteration = 123

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)
    mock_logger.clearml_logger.report_scalar.assert_called_once_with(iteration=123, series="0", title="lr", value=0.01)

    wrapper = OptimizerParamsHandler(optimizer, param_name="lr", tag="generator")
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)
    mock_logger.clearml_logger.report_scalar.assert_called_once_with(
        iteration=123, series="0", title="generator/lr", value=0.01
    )


def test_output_handler_with_wrong_logger_type():
    wrapper = OutputHandler("tag", output_transform=lambda x: x)

    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler OutputHandler works only with ClearMLLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_output_handler_output_transform(dirname):
    wrapper = OutputHandler("tag", output_transform=lambda x: x)
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.output = 12345
    mock_engine.state.iteration = 123

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    mock_logger.clearml_logger.report_scalar.assert_called_once_with(
        iteration=123, series="output", title="tag", value=12345
    )

    wrapper = OutputHandler("another_tag", output_transform=lambda x: {"loss": x})
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)
    mock_logger.clearml_logger.report_scalar.assert_called_once_with(
        iteration=123, series="loss", title="another_tag", value=12345
    )


def test_output_handler_metric_names(dirname):
    wrapper = OutputHandler("tag", metric_names=["a", "b"])
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.clearml_logger.report_scalar.call_count == 2
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="tag", series="a", iteration=5, value=12.23),
            call(title="tag", series="b", iteration=5, value=23.45),
        ],
        any_order=True,
    )

    wrapper = OutputHandler("tag", metric_names=["a", "c"])

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 55.56, "c": "Some text"})
    mock_engine.state.iteration = 7

    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    with pytest.warns(UserWarning, match=r"Logger output_handler can not log metrics value type"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.clearml_logger.report_scalar.call_count == 1
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [call(title="tag", series="a", iteration=7, value=55.56)], any_order=True
    )

    # all metrics
    wrapper = OutputHandler("tag", metric_names="all")
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.clearml_logger.report_scalar.call_count == 2
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="tag", series="a", iteration=5, value=12.23),
            call(title="tag", series="b", iteration=5, value=23.45),
        ],
        any_order=True,
    )

    # log a torch vector
    wrapper = OutputHandler("tag", metric_names="all")
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    vector = torch.tensor([0.1, 0.2, 0.1, 0.2, 0.33])
    mock_engine.state = State(metrics={"vector": vector})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.clearml_logger.report_scalar.call_count == 5
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [call(title="tag/vector", series=str(i), iteration=5, value=vector[i].item()) for i in range(5)],
        any_order=True,
    )

    # log a torch tensor (ndimension = 0)
    wrapper = OutputHandler("tag", metric_names="all")
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": torch.tensor(12.23), "b": torch.tensor(23.45), "c": torch.tensor(5.01)})
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.clearml_logger.report_scalar.call_count == 3
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="tag", series="a", iteration=5, value=torch.tensor(12.23).item()),
            call(title="tag", series="b", iteration=5, value=torch.tensor(23.45).item()),
            call(title="tag", series="c", iteration=5, value=torch.tensor(5.01).item()),
        ],
        any_order=True,
    )

    wrapper = OutputHandler("tag", metric_names="all")
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State(
        metrics={
            "a": 123,
            "b": {"c": [2.34, {"d": 1}]},
            "c": (22, [33, -5.5], {"e": 32.1}),
        }
    )
    mock_engine.state.iteration = 5

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.clearml_logger.report_scalar.call_count == 7
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="tag", series="a", iteration=5, value=123),
            call(title="tag/b", series="c/0", iteration=5, value=2.34),
            call(title="tag/b", series="c/1/d", iteration=5, value=1),
            call(title="tag/c", series="0", iteration=5, value=22),
            call(title="tag/c", series="1/0", iteration=5, value=33),
            call(title="tag/c", series="1/1", iteration=5, value=-5.5),
            call(title="tag/c", series="2/e", iteration=5, value=32.1),
        ],
        any_order=True,
    )


def test_output_handler_both(dirname):
    wrapper = OutputHandler("tag", metric_names=["a", "b"], output_transform=lambda x: {"loss": x})
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State(metrics={"a": 12.23, "b": 23.45})
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    assert mock_logger.clearml_logger.report_scalar.call_count == 3
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="tag", series="a", iteration=5, value=12.23),
            call(title="tag", series="b", iteration=5, value=23.45),
            call(title="tag", series="loss", iteration=5, value=12345),
        ],
        any_order=True,
    )


def test_output_handler_with_wrong_global_step_transform_output():
    def global_step_transform(*args, **kwargs):
        return "a"

    wrapper = OutputHandler("tag", output_transform=lambda x: {"loss": x}, global_step_transform=global_step_transform)
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    with pytest.raises(TypeError, match="global_step must be int"):
        wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)


def test_output_handler_with_global_step_from_engine():
    mock_another_engine = MagicMock()
    mock_another_engine.state = State()
    mock_another_engine.state.epoch = 10
    mock_another_engine.state.output = 12.345

    wrapper = OutputHandler(
        "tag",
        output_transform=lambda x: {"loss": x},
        global_step_transform=global_step_from_engine(mock_another_engine),
    )

    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 1
    mock_engine.state.output = 0.123

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    assert mock_logger.clearml_logger.report_scalar.call_count == 1
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [call(title="tag", series="loss", iteration=mock_another_engine.state.epoch, value=mock_engine.state.output)]
    )

    mock_another_engine.state.epoch = 11
    mock_engine.state.output = 1.123

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    assert mock_logger.clearml_logger.report_scalar.call_count == 2
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [call(title="tag", series="loss", iteration=mock_another_engine.state.epoch, value=mock_engine.state.output)]
    )


def test_output_handler_state_attrs():
    wrapper = OutputHandler("tag", state_attributes=["alpha", "beta", "gamma"])
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.iteration = 5
    mock_engine.state.alpha = 3.899
    mock_engine.state.beta = torch.tensor(12.0)
    mock_engine.state.gamma = torch.tensor([21.0, 6.0])

    wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)

    assert mock_logger.clearml_logger.report_scalar.call_count == 4
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="tag", series="alpha", iteration=5, value=3.899),
            call(title="tag", series="beta", iteration=5, value=12.0),
            call(title="tag/gamma", series="0", iteration=5, value=21.0),
            call(title="tag/gamma", series="1", iteration=5, value=6.0),
        ],
        any_order=True,
    )


def test_output_handler_with_global_step_transform():
    def global_step_transform(*args, **kwargs):
        return 10

    wrapper = OutputHandler("tag", output_transform=lambda x: {"loss": x}, global_step_transform=global_step_transform)
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5
    mock_engine.state.output = 12345

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    assert mock_logger.clearml_logger.report_scalar.call_count == 1
    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [call(title="tag", series="loss", iteration=10, value=12345)]
    )


def test_weights_scalar_handler_wrong_setup():
    model = MagicMock(spec=torch.nn.Module)
    wrapper = WeightsScalarHandler(model)
    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler WeightsScalarHandler works only with ClearMLLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_weights_scalar_handler(dummy_model_factory):
    model = dummy_model_factory(with_grads=True, with_frozen_layer=False)

    # define test wrapper to test with and without optional tag
    def _test(tag=None):
        wrapper = WeightsScalarHandler(model, tag=tag)
        mock_logger = MagicMock(spec=ClearMLLogger)
        mock_logger.clearml_logger = MagicMock()

        mock_engine = MagicMock()
        mock_engine.state = State()
        mock_engine.state.epoch = 5

        wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

        tag_prefix = f"{tag}/" if tag else ""

        assert mock_logger.clearml_logger.report_scalar.call_count == 4
        mock_logger.clearml_logger.report_scalar.assert_has_calls(
            [
                call(title=tag_prefix + "weights_norm/fc1", series="weight", iteration=5, value=0.0),
                call(title=tag_prefix + "weights_norm/fc1", series="bias", iteration=5, value=0.0),
                call(title=tag_prefix + "weights_norm/fc2", series="weight", iteration=5, value=12.0),
                call(title=tag_prefix + "weights_norm/fc2", series="bias", iteration=5, value=math.sqrt(12.0)),
            ],
            any_order=True,
        )

    _test()
    _test(tag="tag")


def test_weights_scalar_handler_whitelist(dummy_model_factory):
    model = dummy_model_factory()

    wrapper = WeightsScalarHandler(model, whitelist=["fc2.weight"])
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    mock_logger.clearml_logger.report_scalar.assert_called_once_with(
        title="weights_norm/fc2", value=ANY, series="weight", iteration=mock_engine.state.epoch
    )
    mock_logger.clearml_logger.report_scalar.reset_mock()

    wrapper = WeightsScalarHandler(model, tag="model", whitelist=["fc1"])
    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="model/weights_norm/fc1", value=ANY, series="weight", iteration=mock_engine.state.epoch),
            call(title="model/weights_norm/fc1", value=ANY, series="bias", iteration=mock_engine.state.epoch),
        ],
        any_order=True,
    )
    assert mock_logger.clearml_logger.report_scalar.call_count == 2
    mock_logger.clearml_logger.report_scalar.reset_mock()

    def weight_selector(n, _):
        return "bias" in n

    wrapper = WeightsScalarHandler(model, tag="model", whitelist=weight_selector)
    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="model/weights_norm/fc1", value=ANY, series="bias", iteration=mock_engine.state.epoch),
            call(title="model/weights_norm/fc2", value=ANY, series="bias", iteration=mock_engine.state.epoch),
        ],
        any_order=True,
    )
    assert mock_logger.clearml_logger.report_scalar.call_count == 2


def test_weights_hist_handler_wrong_setup():
    model = MagicMock(spec=torch.nn.Module)
    wrapper = WeightsHistHandler(model)
    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler 'WeightsHistHandler' works only with ClearMLLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_weights_hist_handler(dummy_model_factory):
    model = dummy_model_factory(with_grads=True, with_frozen_layer=False)

    # define test wrapper to test with and without optional tag
    def _test(tag=None):
        wrapper = WeightsHistHandler(model, tag=tag)
        mock_logger = MagicMock(spec=ClearMLLogger)
        mock_logger.grad_helper = MagicMock()

        mock_engine = MagicMock()
        mock_engine.state = State()
        mock_engine.state.epoch = 5

        wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

        tag_prefix = f"{tag}/" if tag else ""

        assert mock_logger.grad_helper.add_histogram.call_count == 4
        mock_logger.grad_helper.add_histogram.assert_has_calls(
            [
                call(title=tag_prefix + "weights_fc1", hist_data=ANY, series="weight", step=5),
                call(title=tag_prefix + "weights_fc1", hist_data=ANY, series="bias", step=5),
                call(title=tag_prefix + "weights_fc2", hist_data=ANY, series="weight", step=5),
                call(title=tag_prefix + "weights_fc2", hist_data=ANY, series="bias", step=5),
            ],
            any_order=True,
        )

    _test()
    _test(tag="tag")


def test_weights_hist_handler_whitelist(dummy_model_factory):
    model = dummy_model_factory()

    wrapper = WeightsHistHandler(model, whitelist=["fc2.weight"])
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.grad_helper = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    mock_logger.grad_helper.add_histogram.assert_called_once_with(
        title="weights_fc2", hist_data=ANY, series="weight", step=5
    )
    mock_logger.grad_helper.add_histogram.reset_mock()

    wrapper = WeightsHistHandler(model, tag="model", whitelist=["fc1"])
    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    mock_logger.grad_helper.add_histogram.assert_has_calls(
        [
            call(title="model/weights_fc1", hist_data=ANY, series="weight", step=5),
            call(title="model/weights_fc1", hist_data=ANY, series="bias", step=5),
        ],
        any_order=True,
    )
    assert mock_logger.grad_helper.add_histogram.call_count == 2
    mock_logger.grad_helper.add_histogram.reset_mock()

    def weight_selector(n, _):
        return "bias" in n

    wrapper = WeightsHistHandler(model, tag="model", whitelist=weight_selector)
    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    mock_logger.grad_helper.add_histogram.assert_has_calls(
        [
            call(title="model/weights_fc1", hist_data=ANY, series="bias", step=5),
            call(title="model/weights_fc2", hist_data=ANY, series="bias", step=5),
        ],
        any_order=True,
    )
    assert mock_logger.grad_helper.add_histogram.call_count == 2


def test_grads_scalar_handler_wrong_setup():
    model = MagicMock(spec=torch.nn.Module)
    wrapper = GradsScalarHandler(model)
    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler GradsScalarHandler works only with ClearMLLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_grads_scalar_handler(dummy_model_factory, norm_mock):
    model = dummy_model_factory(with_grads=True, with_frozen_layer=False)

    # define test wrapper to test with and without optional tag
    def _test(tag=None):
        wrapper = GradsScalarHandler(model, reduction=norm_mock, tag=tag)
        mock_logger = MagicMock(spec=ClearMLLogger)
        mock_logger.clearml_logger = MagicMock()

        mock_engine = MagicMock()
        mock_engine.state = State()
        mock_engine.state.epoch = 5
        norm_mock.reset_mock()

        wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

        tag_prefix = f"{tag}/" if tag else ""

        mock_logger.clearml_logger.report_scalar.assert_has_calls(
            [
                call(
                    title=tag_prefix + "grads_norm/fc1", value=ANY, series="weight", iteration=mock_engine.state.epoch
                ),
                call(title=tag_prefix + "grads_norm/fc1", value=ANY, series="bias", iteration=mock_engine.state.epoch),
                call(
                    title=tag_prefix + "grads_norm/fc2", value=ANY, series="weight", iteration=mock_engine.state.epoch
                ),
                call(title=tag_prefix + "grads_norm/fc2", value=ANY, series="bias", iteration=mock_engine.state.epoch),
            ],
            any_order=True,
        )
        assert mock_logger.clearml_logger.report_scalar.call_count == 4
        assert norm_mock.call_count == 4

    _test()
    _test(tag="tag")


def test_grads_scalar_handler_whitelist(dummy_model_factory):
    model = dummy_model_factory()

    wrapper = GradsScalarHandler(model, whitelist=["fc2.weight"])
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.clearml_logger = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    mock_logger.clearml_logger.report_scalar.assert_called_once_with(
        title="grads_norm/fc2", value=ANY, series="weight", iteration=mock_engine.state.epoch
    )
    mock_logger.clearml_logger.report_scalar.reset_mock()

    wrapper = GradsScalarHandler(model, tag="model", whitelist=["fc1"])
    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="model/grads_norm/fc1", value=ANY, series="weight", iteration=mock_engine.state.epoch),
            call(title="model/grads_norm/fc1", value=ANY, series="bias", iteration=mock_engine.state.epoch),
        ],
        any_order=True,
    )
    assert mock_logger.clearml_logger.report_scalar.call_count == 2
    mock_logger.clearml_logger.report_scalar.reset_mock()

    def weight_selector(n, _):
        return "bias" in n

    wrapper = GradsScalarHandler(model, tag="model", whitelist=weight_selector)
    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    mock_logger.clearml_logger.report_scalar.assert_has_calls(
        [
            call(title="model/grads_norm/fc1", value=ANY, series="bias", iteration=mock_engine.state.epoch),
            call(title="model/grads_norm/fc2", value=ANY, series="bias", iteration=mock_engine.state.epoch),
        ],
        any_order=True,
    )
    assert mock_logger.clearml_logger.report_scalar.call_count == 2


def test_grads_hist_handler_wrong_setup():
    model = MagicMock(spec=torch.nn.Module)
    wrapper = GradsHistHandler(model)
    mock_logger = MagicMock()
    mock_engine = MagicMock()
    with pytest.raises(RuntimeError, match="Handler 'GradsHistHandler' works only with ClearMLLogger"):
        wrapper(mock_engine, mock_logger, Events.ITERATION_STARTED)


def test_grads_hist_handler(dummy_model_factory):
    model = dummy_model_factory(with_grads=True, with_frozen_layer=False)

    # define test wrapper to test with and without optional tag
    def _test(tag=None):
        wrapper = GradsHistHandler(model, tag=tag)
        mock_logger = MagicMock(spec=ClearMLLogger)
        mock_logger.grad_helper = MagicMock()

        mock_engine = MagicMock()
        mock_engine.state = State()
        mock_engine.state.epoch = 5

        wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

        tag_prefix = f"{tag}/" if tag else ""

        assert mock_logger.grad_helper.add_histogram.call_count == 4
        mock_logger.grad_helper.add_histogram.assert_has_calls(
            [
                call(title=tag_prefix + "grads_fc1", hist_data=ANY, series="weight", step=5),
                call(title=tag_prefix + "grads_fc1", hist_data=ANY, series="bias", step=5),
                call(title=tag_prefix + "grads_fc2", hist_data=ANY, series="weight", step=5),
                call(title=tag_prefix + "grads_fc2", hist_data=ANY, series="bias", step=5),
            ],
            any_order=True,
        )

    _test()
    _test(tag="tag")


def test_grads_hist_handler_whitelist(dummy_model_factory):
    model = dummy_model_factory()

    wrapper = GradsHistHandler(model, whitelist=["fc2.weight"])
    mock_logger = MagicMock(spec=ClearMLLogger)
    mock_logger.grad_helper = MagicMock()

    mock_engine = MagicMock()
    mock_engine.state = State()
    mock_engine.state.epoch = 5

    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)
    mock_logger.grad_helper.add_histogram.assert_called_once_with(
        title="grads_fc2", hist_data=ANY, series="weight", step=5
    )
    mock_logger.grad_helper.reset_mock()

    wrapper = GradsHistHandler(model, tag="model", whitelist=["fc1"])
    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    mock_logger.grad_helper.add_histogram.assert_has_calls(
        [
            call(title="model/grads_fc1", hist_data=ANY, series="weight", step=5),
            call(title="model/grads_fc1", hist_data=ANY, series="bias", step=5),
        ],
        any_order=True,
    )
    assert mock_logger.grad_helper.add_histogram.call_count == 2
    mock_logger.grad_helper.reset_mock()

    def weight_selector(n, _):
        return "bias" in n

    wrapper = GradsHistHandler(model, tag="model", whitelist=weight_selector)
    wrapper(mock_engine, mock_logger, Events.EPOCH_STARTED)

    mock_logger.grad_helper.add_histogram.assert_has_calls(
        [
            call(title="model/grads_fc1", hist_data=ANY, series="bias", step=5),
            call(title="model/grads_fc2", hist_data=ANY, series="bias", step=5),
        ],
        any_order=True,
    )
    assert mock_logger.grad_helper.add_histogram.call_count == 2


def test_integration(dirname):
    n_epochs = 5
    data = list(range(50))

    losses = torch.rand(n_epochs * len(data))
    losses_iter = iter(losses)

    def update_fn(engine, batch):
        return next(losses_iter)

    trainer = Engine(update_fn)

    with pytest.warns(UserWarning, match="ClearMLSaver: running in bypass mode"):
        ClearMLLogger.set_bypass_mode(True)
        logger = ClearMLLogger(output_uri=dirname)

        def dummy_handler(engine, logger, event_name):
            global_step = engine.state.get_event_attrib_value(event_name)
            test_value = 0.3  # example
            logger.clearml_logger.report_scalar(title="", series="", value=test_value, iteration=global_step)

        logger.attach(trainer, log_handler=dummy_handler, event_name=Events.EPOCH_COMPLETED)

        trainer.run(data, max_epochs=n_epochs)
        logger.close()


def test_integration_as_context_manager(dirname):
    n_epochs = 5
    data = list(range(50))

    losses = torch.rand(n_epochs * len(data))
    losses_iter = iter(losses)

    def update_fn(engine, batch):
        return next(losses_iter)

    with pytest.warns(UserWarning, match="ClearMLSaver: running in bypass mode"):
        ClearMLLogger.set_bypass_mode(True)
        with ClearMLLogger(output_uri=dirname) as clearml_logger:
            trainer = Engine(update_fn)

            def dummy_handler(engine, logger, event_name):
                global_step = engine.state.get_event_attrib_value(event_name)
                test_value = 0.3  # example
                logger.clearml_logger.report_scalar(title="", series="", value=test_value, iteration=global_step)

            clearml_logger.attach(trainer, log_handler=dummy_handler, event_name=Events.EPOCH_COMPLETED)

            trainer.run(data, max_epochs=n_epochs)


def test_clearml_logger_getattr_method(dirname):
    with pytest.warns(UserWarning, match="ClearMLSaver: running in bypass mode"):
        ClearMLLogger.set_bypass_mode(True)

        logger = ClearMLLogger(output_uri=dirname)

        # Create a mock clearml.Logger() object
        mock_logger = MagicMock()
        logger.clearml_logger = mock_logger

        # Test a method called by __getattr__ calls the corresponding method of the mock project.
        logger.report_single_value("accuracy", 0.72)
        mock_logger.report_single_value.assert_called_once_with("accuracy", 0.72)

        # Test a method called by __getattr__ calls the corresponding classmethod of the mock project's class.
        logger.current_logger()
        mock_logger.current_logger.assert_called_once()

        logger.close()


def test_clearml_logger_get_task_bypass(dirname):
    with pytest.warns(UserWarning, match="ClearMLSaver: running in bypass mode"):
        ClearMLLogger.set_bypass_mode(True)

        with ClearMLLogger(output_uri=dirname) as clearml_logger:
            task = clearml_logger.get_task()
            assert isinstance(task, clearml.Task)
            assert task == clearml.Task.current_task()
            task.close()


def test_clearml_disk_saver_integration():
    model = torch.nn.Module()
    to_save_serializable = {"model": model}
    with pytest.warns(UserWarning, match="ClearMLSaver created a temporary checkpoints directory"):
        mock_logger = MagicMock(spec=ClearMLLogger)
        clearml.Task.current_task = MagicMock(spec=clearml.Task)
        clearml_saver = ClearMLSaver(mock_logger)
        clearml.binding.frameworks.WeightsFileHandler.create_output_model = MagicMock()

    checkpoint = Checkpoint(to_save=to_save_serializable, save_handler=clearml_saver, n_saved=1)

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)
    checkpoint(trainer)
    trainer.state.iteration = 1
    checkpoint(trainer)
    if clearml_saver._atomic:
        assert clearml.binding.frameworks.WeightsFileHandler.create_output_model.call_count == 2
    else:
        saved_files = list(os.listdir(clearml_saver.dirname))
        assert len(saved_files) == 1
        assert saved_files[0] == "model_1.pt"


def test_clearml_disk_saver_integration_no_logger():
    model = torch.nn.Module()
    to_save_serializable = {"model": model}

    with pytest.warns(UserWarning, match="ClearMLSaver created a temporary checkpoints directory"):
        clearml.Task.current_task = MagicMock(spec=clearml.Task)
        clearml.binding.frameworks.WeightsFileHandler.create_output_model = MagicMock()
        clearml_saver = ClearMLSaver()
        checkpoint = Checkpoint(to_save=to_save_serializable, save_handler=clearml_saver, n_saved=1)

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)
    checkpoint(trainer)
    trainer.state.iteration = 1
    checkpoint(trainer)

    if clearml_saver._atomic:
        assert clearml.binding.frameworks.WeightsFileHandler.create_output_model.call_count == 2
    else:
        saved_files = list(os.listdir(clearml_saver.dirname))
        assert len(saved_files) == 1
        assert saved_files[0] == "model_1.pt"


def test_clearml_saver_callbacks():
    mock_task = MagicMock(spec=clearml.Task)
    mock_task.name = "check-task"

    mock_model = MagicMock(spec=clearml.OutputModel)

    model_info = WeightsFileHandler.ModelInfo(
        model=mock_model,
        upload_filename="test.pt",
        local_model_path="",
        local_model_id="",
        framework=Framework.pytorch,
        task=mock_task,
    )

    mock_model_info = MagicMock(spec_set=model_info)

    # Simulate 4 calls to save model and 2 to remove (n_saved=2)
    filenames = [
        "best_model_5_val_acc=0.123.pt",
        "best_model_6_val_acc=0.234.pt",
        "best_model_7_val_acc=0.356.pt",
        "best_model_8_val_acc=0.456.pt",
    ]
    metadata_list = [
        {"basename": "best_model", "score_name": "val_acc", "priority": 0.123},
        {"basename": "best_model", "score_name": "val_acc", "priority": 0.234},
        {"basename": "best_model", "score_name": "val_acc", "priority": 0.345},
        {"basename": "best_model", "score_name": "val_acc", "priority": 0.456},
    ]
    dirname = "/tmp/test"

    _checkpoint_slots = defaultdict(list)

    n_saved = 2

    for i, (filename, metadata) in enumerate(zip(filenames, metadata_list)):
        mock_model_info.upload_filename = filename

        if i >= n_saved:
            # Remove
            filename_to_remove = filenames[i % n_saved]
            for slots in _checkpoint_slots.values():
                try:
                    slots[slots.index(filename_to_remove)] = None
                except ValueError:
                    pass
                else:
                    i = i % n_saved
                    break

        basename = metadata["basename"]
        checkpoint_key = (dirname, basename)

        context = ClearMLSaver._CallbacksContext(
            callback_type=WeightsFileHandler.CallbackType,
            slots=_checkpoint_slots[checkpoint_key],
            checkpoint_key=str(checkpoint_key),
            filename=filename,
            basename=basename,
            metadata=metadata,
        )

        output_model_info = context.pre_callback(str(WeightsFileHandler.CallbackType.save), mock_model_info)
        assert (
            hasattr(output_model_info, "upload_filename") and f"{basename}_{i}.pt" in output_model_info.upload_filename
        )
        assert hasattr(output_model_info, "local_model_id") and str(checkpoint_key) in output_model_info.local_model_id

        output_model_info = context.post_callback(str(WeightsFileHandler.CallbackType.save), mock_model_info)
        assert hasattr(output_model_info, "model") and hasattr(output_model_info.model, "name")
        assert hasattr(output_model_info, "model") and hasattr(output_model_info.model, "comment")
        assert isinstance(output_model_info.model.name, str) and filename in output_model_info.model.name
        assert (
            isinstance(output_model_info.model.comment, str)
            and metadata["basename"] in output_model_info.model.comment
            and metadata["score_name"] in output_model_info.model.comment
        )


class DummyModel(torch.nn.Module):
    def __init__(self):
        super(DummyModel, self).__init__()
        self.net = torch.nn.Linear(2, 2)

    def forward(self, x):
        return self.net(x)


def _test_save_model_optimizer_lr_scheduler_with_state_dict(device, on_zero_rank=False):
    if idist.get_rank() == 0:
        clearml.Task.current_task = MagicMock(spec=clearml.Task)
        clearml.binding.frameworks.WeightsFileHandler.create_output_model = MagicMock()

    torch.manual_seed(23)

    model = DummyModel().to(device)

    optim = torch.optim.SGD(model.parameters(), lr=0.1)
    lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optim, gamma=0.5)

    def update_fn(engine, batch):
        x = torch.rand((4, 2)).to(device)
        optim.zero_grad()
        y = model(x)
        # Below code raises: RuntimeError: torch_xla/csrc/tensor_impl.cpp:144 : XLA tensors do not have storage
        # Probably related to https://github.com/pytorch/xla/issues/2576
        # loss = y.pow(2.0).sum()
        loss = y.sum()
        loss.backward()
        if idist.has_xla_support:
            import torch_xla.core.xla_model as xm

            xm.optimizer_step(optim, barrier=True)
        else:
            optim.step()
        lr_scheduler.step()

    engine = Engine(update_fn)

    to_save = {"model": model, "optimizer": optim, "lr_scheduler": lr_scheduler}

    with pytest.warns(UserWarning, match=r"ClearMLSaver created a temporary checkpoints directory"):
        clearml_saver = ClearMLSaver()

    if (not on_zero_rank) or (on_zero_rank and idist.get_rank() == 0):
        checkpoint = Checkpoint(to_save=to_save, save_handler=clearml_saver, n_saved=1)
        engine.add_event_handler(Events.EPOCH_COMPLETED, checkpoint)

    engine.run([0], max_epochs=4)

    idist.barrier()

    saved_objects = sorted(os.listdir(clearml_saver.dirname))
    # saved object is ['PREFIX_checkpoint_3.pt', ]
    saved_checkpoint = clearml_saver.dirname / saved_objects[0]

    if idist.has_xla_support:
        device = "cpu"

    loaded_obj = torch.load(saved_checkpoint, map_location=device)
    for f in ["model", "optimizer", "lr_scheduler"]:
        assert f in loaded_obj
    loaded_model_state_dict = loaded_obj["model"]
    loaded_optimizer_state_dict = loaded_obj["optimizer"]
    loaded_lr_scheduler_state_dict = loaded_obj["lr_scheduler"]

    assert isinstance(loaded_model_state_dict, dict)
    assert isinstance(loaded_optimizer_state_dict, dict)
    assert isinstance(loaded_lr_scheduler_state_dict, dict)

    # Specifically move device to CPU first
    model_state_dict = model.cpu().state_dict()
    for key in model_state_dict.keys():
        assert key in loaded_model_state_dict
        model_value = model_state_dict[key]
        loaded_model_value = loaded_model_state_dict[key]
        assert (model_value.cpu().numpy() == loaded_model_value.cpu().numpy()).all()

    optim_state_dict = optim.state_dict()
    for key in optim_state_dict.keys():
        assert key in loaded_optimizer_state_dict
        optim_value = optim_state_dict[key]
        loaded_optim_value = loaded_optimizer_state_dict[key]
        if idist.get_rank() == 0:
            assert optim_value == loaded_optim_value

    lr_scheduler_state_dict = lr_scheduler.state_dict()
    for key in lr_scheduler_state_dict.keys():
        assert key in loaded_lr_scheduler_state_dict
        lr_scheduler_value = lr_scheduler_state_dict[key]
        loaded_lr_scheduler_value = loaded_lr_scheduler_state_dict[key]
        assert lr_scheduler_value == loaded_lr_scheduler_value


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):
    device = idist.device()
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device)
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device, on_zero_rank=True)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl):
    device = idist.device()
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device)
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device, on_zero_rank=True)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Not on TPU device")
def test_distrib_single_device_xla():
    device = idist.device()
    assert "xla" in device.type
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device)


def _test_save_model_optimizer_lr_scheduler_with_state_dict_xla_nprocs(index):
    device = idist.device()
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device)

    import time

    # hack to have all proc properly sync:
    time.sleep(1)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Not on TPU device")
def test_distrib_single_device_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_save_model_optimizer_lr_scheduler_with_state_dict_xla_nprocs, args=(), nprocs=n)
